# -*- coding: utf-8 -*-

"""Setup and run MOPAC"""

import calendar
import configparser
import csv
from datetime import datetime, timezone
import importlib
import logging
import os
import os.path
from pathlib import Path
import platform
import pprint
import shutil
import string
import time

from cpuinfo import get_cpu_info

import molsystem
import seamm
import seamm_exec
from seamm_util import Configuration
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __
import mopac_step

logger = logging.getLogger(__name__)
job = printing.getPrinter()
printer = printing.getPrinter("mopac")

# Add MOPAC's properties to the standard properties
resources = importlib.resources.files("mopac_step") / "data"
csv_file = resources / "properties.csv"
molsystem.add_properties_from_file(csv_file)


class MOPAC(mopac_step.MOPACBase):
    def __init__(
        self,
        flowchart=None,
        namespace="org.molssi.seamm.mopac",
        extension=None,
        title="MOPAC",
        logger=logger,
    ):
        """Initialize the node"""

        logger.debug("Creating MOPAC {}".format(self))

        # Create the subflowchart and proceed
        if title == "MOPAC":
            self.subflowchart = seamm.Flowchart(
                name="MOPAC",
                parent=self,
                namespace=namespace,
                directory=flowchart.root_directory,
            )
        self._data = {}
        self._lattice_opt = True
        self._lattice_shear = True
        self._lattice_couple = "none"
        self._input_only = False

        super().__init__(
            flowchart=flowchart, title=title, extension=extension, logger=logger
        )

        # Set up the timing information
        self._timing_data = []
        self._timing_path = Path("~/.seamm.d/timing/mopac.csv").expanduser()
        self._timing_header = [
            "node",  # 0
            "cpu",  # 1
            "cpu_version",  # 2
            "cpu_count",  # 3
            "cpu_speed",  # 4
            "date",  # 5
            "H_SMILES",  # 6
            "ISOMERIC_SMILES",  # 7
            "formula",  # 8
            "net_charge",  # 9
            "spin_multiplicity",  # 10
            "keywords",  # 11
            "nproc",  # 12
            "time",  # 13
        ]
        try:
            self._timing_path.parent.mkdir(parents=True, exist_ok=True)

            self._timing_data = 14 * [""]
            self._timing_data[0] = platform.node()
            tmp = get_cpu_info()
            if "arch" in tmp:
                self._timing_data[1] = tmp["arch"]
            if "cpuinfo_version_string" in tmp:
                self._timing_data[2] = tmp["cpuinfo_version_string"]
            if "count" in tmp:
                self._timing_data[3] = str(tmp["count"])
            if "hz_advertized_friendly" in tmp:
                self._timing_data[4] = tmp["hz_advertized_friendly"]

            if not self._timing_path.exists():
                with self._timing_path.open("w", newline="") as fd:
                    writer = csv.writer(fd)
                    writer.writerow(self._timing_header)
        except Exception:
            self._timing_data = None

    @property
    def input_only(self):
        """Whether to write the input only, not run MOPAC."""
        return self._input_only

    @input_only.setter
    def input_only(self, value):
        self._input_only = value

    def description_text(self, P=None):
        """Return a short description of this step.

        Return a nicely formatted string describing what this step will
        do.

        Keyword arguments:
            P: a dictionary of parameter values, which may be variables
                or final values. If None, then the parameters values will
                be used as is.
        """
        # Work through children. Get the first real node
        node = self.subflowchart.get_node("1").next()

        text = self.header + "\n\n"
        while node is not None:
            text += __(node.description_text(), indent=4 * " ").__str__()
            text += "\n"
            node = node.next()

        return text

    def run(self, printer=printer):
        """Run MOPAC"""
        # Create the directory
        directory = Path(self.directory)
        directory.mkdir(parents=True, exist_ok=True)

        next_node = super().run(printer)

        system, configuration = self.get_system_configuration(None)
        n_atoms = configuration.n_atoms
        if n_atoms == 0:
            self.logger.error("MOPAC run(): there is no structure!")
            raise RuntimeError("MOPAC run(): there is no structure!")

        # Print our header to the main output
        printer.normal(self.header)
        printer.normal("")

        # Access the options
        options = self.options
        seamm_options = self.global_options

        extra_keywords = ["AUX(MOS=10,XP,XS,PRECISION=3)"]

        # Always add the charge since that will cause MOZYME, if used, to check.
        extra_keywords.append(f"CHARGE={configuration.charge}")
        # And the spin multiplicity
        multiplicity = configuration.spin_multiplicity
        if multiplicity <= 10:
            extra_keywords.append(
                (
                    "SINGLET",
                    "DOUBLET",
                    "TRIPLET",
                    "QUARTET",
                    "QUINTET",
                    "SEXTET",
                    "SEPTET",
                    "OCTET",
                    "NONET",
                )[multiplicity - 1]
            )
        else:
            extra_keywords.append(f"MS={(multiplicity - 1) / 2}")

        n_active_electrons = configuration.n_active_electrons
        n_active_orbitals = configuration.n_active_orbitals

        if n_active_orbitals > 0:
            extra_keywords.append(f"OPEN({n_active_electrons},{n_active_orbitals})")
            state = configuration.state
            if state != "1":
                extra_keywords.append(f"ROOT={state}")
        else:
            if multiplicity > 1:
                extra_keywords.append("UHF")

        # All Lanthanides (except La and Lu) must use the SPARKLES keyword.
        # La and Lu use the SPARKLES keyword optionally, depending
        # if you're looking for good structure (do use SPARKLES) or
        # energy (do not use SPARKLES)
        La = [
            "Ce",
            "Pr",
            "Nd",
            "Pm",
            "Sm",
            "Eu",
            "Gd",
            "Tb",
            "Dy",
            "Ho",
            "Er",
            "Tm",
            "Yb",
        ]

        La_list = set(La) & set(configuration.atoms.symbols)

        if len(La_list) > 0:
            extra_keywords.append("SPARKLES")

        # if mopac_num_threads > 1:
        #     extra_keywords.append("THREADS={}".format(mopac_num_threads))

        # Work through the subflowchart to find out what to do.
        self.subflowchart.root_directory = self.flowchart.root_directory

        # Get the first real node
        node = self.subflowchart.get_node("1").next()

        text = ""
        n_calculations = []
        all_keywords = []
        while node:
            node.parent = self
            inputs = node.get_input()
            n_calculations.append(len(inputs))
            for keywords, structure, comment in inputs:
                lines = []
                if "OLDGEO" not in keywords:
                    structure_lines, symlines = self.mopac_structure()
                    if symlines != "" and "SYMMETRY" not in extra_keywords:
                        extra_keywords.append("SYMMETRY")
                else:
                    symlines = ""
                all_keywords.append(" ".join(keywords + extra_keywords))
                lines.append(" ".join(keywords + extra_keywords))
                lines.append(system.name)
                if comment is None:
                    lines.append(configuration.name)
                else:
                    lines.append(comment)

                text += "\n".join(lines)
                text += "\n"
                if structure is None:
                    if "OLDGEO" not in keywords:
                        text += structure_lines
                        text += "\n"
                        if symlines != "":
                            text += symlines
                            text += "\n"
                else:
                    text += structure_lines
                    text += "\n"
            node = node.next()

        # Check for successful run, don't rerun
        output = ""  # Text output to print
        success = directory / "success.dat"
        if success.exists():
            self._timing_data = None
        else:
            # Input files
            files = {"mopac.dat": text}
            self.logger.debug("mopac.dat:\n" + files["mopac.dat"])
            for filename in files:
                path = directory / filename
                path.write_text(files[filename])

            if self.input_only:
                self._timing_data = None
            else:
                # Get the computational environment and set limits
                ce = seamm_exec.computational_environment()

                n_cores = ce["NTASKS"]
                if seamm_options["ncores"] != "available":
                    n_cores = min(n_cores, int(seamm_options["ncores"]))
                # Currently, on the Mac, it is not clear that any parallelism helps
                # much.

                n_hydrogens = configuration.atoms.get_n_atoms("atno", "==", 1)
                n_basis = (n_atoms - n_hydrogens) * 4 + n_hydrogens
                if options["ncores"] == "default":
                    # Since it is the matrix diagonalization, work out rough
                    # size of matrix

                    # Wild guess!
                    # tmp = max(1, int(pow(n_basis / 1000, 3)))
                    # if tmp < n_cores:
                    #     n_cores = tmp

                    # It appears that MOPAC gets little benefit from parallel,
                    # so run serial
                    tmp = 1
                else:
                    tmp = int(options["ncores"])
                if tmp < n_cores:
                    n_cores = tmp
                if n_cores < 1:
                    n_cores = 1
                ce["NTASKS"] = n_cores

                output = (
                    f"MOPAC will use {n_cores} threads for {n_atoms} atoms with "
                    f"{n_basis} basis functions."
                )
                output = __(output, indent=8 * " ")

                env = {
                    "OMP_NUM_THREADS": str(n_cores),
                }

                executor = self.flowchart.executor

                # Read configuration file for MOPAC if it exists
                executor_type = executor.name
                full_config = configparser.ConfigParser()
                ini_dir = Path(seamm_options["root"]).expanduser()
                path = ini_dir / "mopac.ini"
                # If the config file doesn't exists, get the default
                if not path.exists():
                    resources = importlib.resources.files("mopac_step") / "data"
                    ini_text = (resources / "mopac.ini").read_text()
                    txt_config = Configuration(path)
                    txt_config.from_string(ini_text)

                    # Work out the conda info needed
                    txt_config.set_value("local", "conda", os.environ["CONDA_EXE"])
                    txt_config.set_value("local", "conda-environment", "seamm-mopac")
                    txt_config.save()

                full_config.read(ini_dir / "mopac.ini")

                # Getting desperate! Look for an executable in the path
                if executor_type not in full_config:
                    path = shutil.which("mopac")
                    if path is None:
                        raise RuntimeError(
                            f"No section for '{executor_type}' in MOPAC ini file "
                            f"({ini_dir / 'mopac.ini'}), nor in the defaults, nor "
                            "in the path!"
                        )
                    else:
                        txt_config = Configuration(path)
                        txt_config.add_section(executor_type)
                        txt_config.set_value(executor_type, "installation", "local")
                        txt_config.set_value(executor_type, "code", str(path))
                        txt_config.save()
                        full_config.read(ini_dir / "mopac.ini")

                config = dict(full_config.items(executor_type))

                # Use the matching version of the seamm-mopac image by default.
                config["version"] = self.version

                return_files = [
                    "mopac.arc",
                    "mopac.out",
                    "mopac.aux",
                    "stdout.txt",
                    "stderr.txt",
                ]

                if self._timing_data is not None:
                    try:
                        self._timing_data[6] = configuration.to_smiles(
                            canonical=True, hydrogens=True
                        )
                    except Exception:
                        self._timing_data[6] = ""
                    try:
                        self._timing_data[7] = configuration.isomeric_smiles
                    except Exception:
                        self._timing_data[7] = ""
                    try:
                        self._timing_data[8] = configuration.formula[0]
                    except Exception:
                        self._timing_data[7] = ""
                    try:
                        self._timing_data[9] = str(configuration.charge)
                    except Exception:
                        self._timing_data[9] = ""
                    try:
                        self._timing_data[10] = str(configuration.spin_multiplicity)
                    except Exception:
                        self._timing_data[10] = ""

                    self._timing_data[11] = " && ".join(all_keywords)
                    self._timing_data[5] = datetime.now(timezone.utc).isoformat()

                t0 = time.time_ns()

                result = executor.run(
                    cmd=["{code}", "mopac.dat", ">", "stdout.txt", "2>", "stderr.txt"],
                    config=config,
                    directory=self.directory,
                    files=files,
                    return_files=return_files,
                    in_situ=True,
                    shell=True,
                    env=env,
                )

                t = (time.time_ns() - t0) / 1.0e9
                if self._timing_data is not None:
                    self._timing_data[13] = f"{t:.3f}"
                    self._timing_data[12] = str(n_cores)
                    try:
                        with self._timing_path.open("a", newline="") as fd:
                            writer = csv.writer(fd)
                            writer.writerow(self._timing_data)
                    except Exception:
                        pass

                if not result:
                    self.logger.error("There was an error running MOPAC")
                    return None

                self.logger.debug("\n" + pprint.pformat(result))

                self.logger.debug(
                    "\n\nOutput from MOPAC\n\n" + result["mopac.out"]["data"] + "\n\n"
                )

        # Ran successfully, put out the success file
        success.write_text("success")

        if not self.input_only:
            # Analyze the results
            self.analyze(n_calculations=n_calculations, output=output)

        # Close the reference handler, which should force it to close the
        # connection.
        self.references = None

        return next_node

    def set_id(self, node_id):
        """Set the id for node to a given tuple"""
        # and set our subnodes
        self.subflowchart.set_ids(node_id)

        return super().set_id(node_id)

    def analyze(self, indent="", lines=[], n_calculations=None, output=""):
        """Read the results from MOPAC calculations and analyze them,
        putting key results into variables for subsequent use by
        other stages
        """

        # Split the aux files into sections for each step
        filename = "mopac.aux"
        with open(os.path.join(self.directory, filename), mode="r") as fd:
            lines_aux = fd.read().splitlines()

        # Find the sections in the file corresponding to sub-tasks
        # MOPAC keeps cumulative times, so fix them
        t_total = 0.0
        aux_data = []
        start = 0
        lineno = 0
        section = 0
        for line in lines_aux:
            if "END OF MOPAC FILE" in line or "END OF MOPAC PROGRAM" in line:
                self.logger.debug("\nAUX file section {}".format(section))
                self.logger.debug("------------------")

                tmp_data = self.parse_aux(lines_aux[start:lineno])
                if "CPU_TIME" in tmp_data:
                    tmp = tmp_data["CPU_TIME"]
                    tmp_data["CPU_TIME"] = tmp - t_total
                    t_total = tmp
                aux_data.append(tmp_data)

                self.logger.debug(pprint.pformat(tmp_data, width=170, compact=True))
            lineno += 1
            if "START OF MOPAC FILE" in line:
                section += 1
                start = lineno

        # Split the output file into sections for each step
        filename = "mopac.out"
        with open(os.path.join(self.directory, filename), mode="r") as fd:
            lines = fd.read().splitlines()

        # Find the sections in the file corresponding to sub-tasks
        out = []
        start = 0
        lineno = 0
        for line in lines:
            if "** Cite this program as:" in line or "Digital Object Ident" in line:
                if lineno == 5:
                    continue
                out.append(lines[start : lineno - 5])
                start = lineno - 6
            lineno += 1
        out.append(lines[start:])

        for data in aux_data:
            # Add main citation for MOPAC
            if "MOPAC_VERSION" in data:
                # like MOPAC2016.20.191M
                release, version = data["MOPAC_VERSION"].split(".", maxsplit=1)
                try:
                    t = datetime.strptime(version[0:-1], "%y.%j")
                    year = t.year
                    month = t.month
                    template = string.Template(self._bibliography["Stewart_2016"])
                    month = calendar.month_abbr[int(month)].lower()
                    citation = template.substitute(
                        month=month, version=version, year=year, release=release
                    )
                    self.references.cite(
                        raw=citation,
                        alias="mopac",
                        module="mopac_step",
                        level=1,
                        note="The principle MOPAC citation.",
                    )
                except Exception:
                    self.references.cite(
                        raw=self._bibliography["stewart_james_j_p_2022_6811510"],
                        alias="mopac",
                        module="mopac_step",
                        level=1,
                        note="The principle MOPAC citation.",
                    )
                break

        # Loop through our subnodes. Get the first real node
        node = self.subflowchart.get_node("1").next()
        first = 0
        n_node = 0
        while node:
            # Print the header for the node
            for value in node.description:
                printer.normal(value)
                if output != "":
                    printer.normal(output)
                    printer.normal("")
                    output = ""

            last = first + n_calculations[n_node]
            if last > len(out):
                logger.error("Could not find the MOPAC output for subjob {last + 1}/")
                node.analyze(data_sections=aux_data[first:last], out_sections=[])
            else:
                node.analyze(
                    data_sections=aux_data[first:last], out_sections=out[first:last]
                )
            first = last

            printer.normal("")

            node = node.next()
            n_node += 1

        if n_node > 1 and "CPU_TIME" in aux_data[-1]:
            text = f"MOPAC took a total of {t_total:.2f} s."
            printer.normal(str(__(text, **data, indent=self.indent)))



# -*- coding: utf-8 -*-

"""Setup and run MOPAC for the Lewis structure"""

import calendar
import configparser
import datetime
import json  # noqa: F401
import logging
from pathlib import Path
import pprint
import os
import string
import textwrap

from tabulate import tabulate

import mopac_step
import seamm
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __

# from seamm_util.printing import FormattedText as __


logger = logging.getLogger(__name__)
job = printing.getPrinter()
printer = printing.getPrinter("mopac")


class LewisStructure(mopac_step.MOPACBase):
    def __init__(self, flowchart=None, extension=None):
        """Initialize the node"""

        logger.debug("Creating MOPAC {}".format(self))

        super().__init__(
            flowchart=flowchart,
            title="Lewis Structure",
            extension=extension,
            logger=logger,
        )

        self._calculation = "Lewis structure"
        self._model = None
        self._metadata = {**mopac_step.metadata}
        # Don't want user to change keywords!
        del self._metadata["keywords"]
        self.parameters = mopac_step.LewisStructureParameters()
        self._lattice_opt = False

    def description_text(self, P=None):
        """Return a short description of this step.

        Return a nicely formatted string describing what this step will
        do.

        Keyword arguments:
            P: a dictionary of parameter values, which may be variables
                or final values. If None, then the parameters values will
                be used as is.
        """
        if not P:
            P = self.parameters.values_to_dict()

        if P["use bonds"]:
            text = (
                "Generate and print the Lewis structure and replace any existing bonds "
                "with those from the Lewis structure."
            )
        else:
            text = "Generate and print the Lewis structure."

        return self.header + "\n" + __(text, **P, indent=4 * " ").__str__()

    def run(self):
        """Run MOPAC"""
        system, configuration = self.get_system_configuration(None)
        n_atoms = configuration.n_atoms
        if n_atoms == 0:
            self.logger.error("MOPAC run(): there is no structure!")
            raise RuntimeError("MOPAC run(): there is no structure!")

        # Print our header to the main output
        printer.normal(self.header)
        printer.normal("")

        # Access the options
        seamm_options = self.global_options

        extra_keywords = ["AUX(MOS=10,XP,XS,PRECISION=3)"]

        # All Lanthanides (except La and Lu) must use the SPARKLES keyword.
        # La and Lu use the SPARKLES keyword optionally, depending
        # if you're looking for good structure (do use SPARKLES) or
        # energy (do not use SPARKLES)
        La = [
            "Ce",
            "Pr",
            "Nd",
            "Pm",
            "Sm",
            "Eu",
            "Gd",
            "Tb",
            "Dy",
            "Ho",
            "Er",
            "Tm",
            "Yb",
        ]

        La_list = set(La) & set(configuration.atoms.symbols)

        if len(La_list) > 0:
            extra_keywords.append("SPARKLES")

        # Always add the charge since that will cause MOZYME, if used, to check.
        extra_keywords.append(f"CHARGE={configuration.charge}")
        # Sparkles affect the number of electrons .. each has 3? .. so ignore
        # multiplicity
        if False:
            # if "SPARKLES" not in extra_keywords:
            # And the spin multiplicity
            multiplicity = configuration.spin_multiplicity
            if multiplicity <= 10:
                extra_keywords.append(
                    (
                        "SINGLET",
                        "DOUBLET",
                        "TRIPLET",
                        "QUARTET",
                        "QUINTET",
                        "SEXTET",
                        "SEPTET",
                        "OCTET",
                        "NONET",
                    )[multiplicity - 1]
                )
            else:
                extra_keywords.append(f"MS={(multiplicity - 1) / 2}")

        next_node = super().run(printer)

        text = ""
        lines = []
        lines.append(" ".join(["LEWIS", "LET", "GEO-OK"] + extra_keywords))
        lines.append(system.name)
        lines.append(configuration.name)

        text += "\n".join(lines)
        text += "\n"
        structure_lines, symlines = self.mopac_structure()
        text += structure_lines
        text += "\n"

        files = {"mopac.dat": text}
        self.logger.debug("mopac.dat:\n" + files["mopac.dat"])
        os.makedirs(self.directory, exist_ok=True)

        executor = self.flowchart.executor

        # Read configuration file for MOPAC
        ini_dir = Path(seamm_options["root"]).expanduser()
        full_config = configparser.ConfigParser()
        full_config.read(ini_dir / "mopac.ini")
        executor_type = executor.name
        if executor_type not in full_config:
            raise RuntimeError(
                f"No section for '{executor_type}' in MOPAC ini file "
                f"({ini_dir / 'mopac.ini'})"
            )
        config = dict(full_config.items(executor_type))

        return_files = ["mopac.arc", "mopac.out", "mopac.aux"]
        result = executor.run(
            cmd=["{code}", "mopac.dat", ">", "stdout.txt", "2>", "stderr.txt"],
            config=config,
            directory=self.directory,
            files=files,
            return_files=return_files,
            in_situ=True,
            shell=True,
        )

        if not result:
            self.logger.error("There was an error running MOPAC")
            return None

        self.logger.debug("\n" + pprint.pformat(result))

        self.logger.debug(
            "\n\nOutput from MOPAC\n\n" + result["mopac.out"]["data"] + "\n\n"
        )

        # Analyze the results
        self.analyze()

        # Close the reference handler, which should force it to close the
        # connection.
        self.references = None

        return next_node

    def analyze(self, indent="", lines=[], n_calculations=None):
        """Read the results from the Lewis calculation and process."""
        _, configuration = self.get_system_configuration(None)

        # Split the aux files into sections for each step
        filename = "mopac.aux"
        with open(os.path.join(self.directory, filename), mode="r") as fd:
            lines_aux = fd.read().splitlines()
        data = self.parse_aux(lines_aux[2:-1])

        # Add main citation for MOPAC
        if "MOPAC_VERSION" in data:
            # like MOPAC2016.20.191M
            if data["MOPAC_VERSION"][0:5] == "MOPAC":
                release, version = data["MOPAC_VERSION"].split(".", maxsplit=1)
                t = datetime.datetime.strptime(version[0:-1], "%y.%j")
                year = t.year
                month = t.month
                template = string.Template(self._bibliography["Stewart_2016"])
                month = calendar.month_abbr[int(month)].lower()
                citation = template.substitute(
                    month=month, version=version, year=year, release=release
                )
                self.references.cite(
                    raw=citation,
                    alias="mopac",
                    module="mopac_step",
                    level=1,
                    note="The principle MOPAC citation.",
                )
            else:
                self.references.cite(
                    raw=self._bibliography["stewart_james_j_p_2022_6811510"],
                    alias="mopac",
                    module="mopac_step",
                    level=1,
                    note="The principle MOPAC citation.",
                )
        else:
            self.references.cite(
                raw=self._bibliography["stewart_james_j_p_2022_6811510"],
                alias="mopac",
                module="mopac_step",
                level=1,
                note="The principle MOPAC citation.",
            )

        # Get the output file
        filename = "mopac.out"
        with open(os.path.join(self.directory, filename), mode="r") as fd:
            lines = iter(fd.read().splitlines())

        # Get the parameters used
        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )
        no_error = P["ignore errors"]
        fallback = P["on errors use connectivity"]

        charge = None
        n_atoms = configuration.n_atoms
        point_group = None
        n_sigma_bonds = None
        n_lone_pairs = None
        n_pi_bonds = None
        sum_positive_charges = None
        sum_negative_charges = None
        ions = {}
        neighbors = data["neighbors"] = [[] for i in range(n_atoms)]
        bonds = data["bonds"] = {"i": [], "j": [], "bondorder": []}
        lone_pairs = data["lone pairs"] = [0] * n_atoms
        have_lewis_structure = False
        charge_error = False
        for line in lines:
            line = line.strip()
            if "MOLECULAR POINT GROUP" in line:
                point_group = line.split()[-1]

            if "Ion Atom No.  Type    Charge" in line:
                for line in lines:
                    line = line.strip()
                    if not line:
                        continue
                    if "COMPUTED CHARGE ON SYSTEM" in line:
                        break
                    ion, atom, type_, charge = line.split()
                    ions[int(atom) - 1] = {"charge": int(charge), "type": type_}
            if "COMPUTED CHARGE ON SYSTEM" in line:
                charge = int(line.split()[4].rstrip(","))
                if "THIS IS THE SAME AS THE CHARGE DEFINED" not in line:
                    charge_error = True
            if "SIGMA BONDS" in line:
                n_sigma_bonds = int(line.split()[2])
            if "LONE PAIRS" in line:
                n_lone_pairs = int(line.split()[2])
            if "PI BONDS" in line:
                n_pi_bonds = int(line.split()[2])
            if "SUM OF POSITIVE CHARGES" in line:
                sum_positive_charges = int(line.split()[4])
            if "SUM OF NEGATIVE CHARGES" in line:
                sum_negative_charges = int(line.split()[4])
            if "TOPOGRAPHY OF SYSTEM" in line:
                next(lines)
                next(lines)
                for line in lines:
                    line = line.strip()
                    if line == "":
                        break
                    tmp = line.split()
                    if len(tmp) > 2:
                        i = int(tmp[0]) - 1
                        if i < 0 or i >= n_atoms:
                            raise RuntimeError(
                                f"Problem with atom index in topography {n_atoms=}: "
                                f"{line}"
                            )
                        neighbors[i] = [int(j) - 1 for j in tmp[2:]]

            if line == "Lewis Structure":
                have_lewis_structure = True
                next(lines)
                next(lines)
                tmp_bonds = {}
                for line in lines:
                    if line.strip() == "":
                        break
                    for start in range(1, 4 * 24, 24):
                        tmp = line[start : start + 24].strip().split()
                        if len(tmp) == 3:
                            i = int(tmp[1]) - 1
                            j = int(tmp[2]) - 1
                            if j < i:
                                i, j = j, i
                            key = f"{i}-{j}"
                            if key in tmp_bonds:
                                tmp_bonds[key] += 1
                            else:
                                tmp_bonds[key] = 1
                        elif len(tmp) == 2:
                            i = int(tmp[1]) - 1
                            lone_pairs[i] += 1
                for key, order in tmp_bonds.items():
                    i, j = key.split("-")
                    bonds["i"].append(int(i))
                    bonds["j"].append(int(j))
                    bonds["bondorder"].append(order)

                # The neighbors according to Lewis structure
                lneighbors = [[] for i in range(n_atoms)]
                for i, j in zip(bonds["i"], bonds["j"]):
                    lneighbors[i].append(j)
                    lneighbors[j].append(i)

        if charge is not None:
            data["charge"] = charge
        if point_group is not None:
            data["point group"] = point_group
        if n_sigma_bonds is not None:
            data["n sigma bonds"] = n_sigma_bonds
        if n_lone_pairs is not None:
            data["n lone pairs"] = n_lone_pairs
        if n_pi_bonds is not None:
            data["n pi bonds"] = n_pi_bonds
        if sum_positive_charges is not None:
            data["sum positive charges"] = sum_positive_charges
        if sum_negative_charges is not None:
            data["sum negative charges"] = sum_negative_charges

        self.logger.debug(f"Point group = {point_group}")
        self.logger.debug(f"     Charge = {charge}")
        self.logger.debug("Neighbors")
        self.logger.debug(json.dumps(neighbors, indent=4))
        self.logger.debug("Bonds")
        self.logger.debug(json.dumps(bonds, indent=4))
        self.logger.debug("Lone pairs")
        self.logger.debug(json.dumps(lone_pairs, indent=4))

        # Check the Lewis structure for consistency with neighbors.
        same = False
        if have_lewis_structure:
            same = True
            for i in range(n_atoms):
                if sorted(neighbors[i]) != sorted(lneighbors[i]):
                    same = False
                    break

            if not same:
                if no_error:
                    self.logger.warning(
                        "The Lewis structure yields different connectivity than the "
                        "simple connectivity."
                    )
                elif not fallback:
                    raise RuntimeError(
                        "The Lewis structure yields different connectivity than the "
                        "simple connectivity shows."
                    )

        # Generate the printed output if requested
        text = ""
        if P["atom cutoff"] != "no printing":
            text += f"   Point group symmetry: {point_group}\n"
            text += f"             Net charge: {charge}\n"
            text += f"Sum of positive charges: {sum_positive_charges}\n"
            text += f"Sum of negative charges: {sum_negative_charges}\n"
            text += f"  Number of sigma bonds: {n_sigma_bonds}\n"
            text += f"     Number of pi bonds: {n_pi_bonds}\n"
            text += f"   Number of lone pairs: {n_lone_pairs}\n"
            text += "\n"
        if charge_error:
            if P["adjust charge"]:
                configuration.charge = charge
                text += (
                    "The total charge on the system has been adjusted to "
                    f"{configuration.charge}.\n"
                )
            elif no_error or fallback:
                text += (
                    f"The charge on the system {configuration.charge} is not "
                    f"the same as Lewis structure indicates: {charge}\n"
                )
            else:
                raise RuntimeError(
                    f"The charge on the system {configuration.charge} is not "
                    f"the same as Lewis structure indicates: {charge}"
                )

        if P["atom cutoff"] == "no printing":
            pass
        elif P["atom cutoff"] == "unlimited" or P["atom cutoff"] >= n_atoms:
            symbol = configuration.atoms.symbols
            nmax = 0
            for atoms in neighbors:
                if len(atoms) > nmax:
                    nmax = len(atoms)
            table = {"Atom": [], "El": [], "Lone Pairs": []}
            for i in range(1, nmax + 1):
                table[f"At{i}"] = []
                table[f"El{i}"] = []
            for i, atoms in enumerate(neighbors):
                table["Atom"].append(i + 1)
                table["El"].append(symbol[i])
                table["Lone Pairs"].append(lone_pairs[i])
                for c, j in enumerate(atoms, start=1):
                    table[f"At{c}"].append(j + 1)
                    table[f"El{c}"].append(symbol[j])
                for c in range(len(atoms) + 1, nmax + 1):
                    table[f"At{c}"].append("")
                    table[f"El{c}"].append("")
            tmp = tabulate(table, headers="keys", tablefmt="pretty")
            length = len(tmp.splitlines()[0])
            text += "\n"
            text += "Atoms bonded to each atom".center(length)
            text += "\n"
            text += tmp
            text += "\n"
            text += "\n"

            table = {
                "At1": [i + 1 for i in bonds["i"]],
                "At2": [j + 1 for j in bonds["j"]],
            }
            txt = table["Bond"] = []
            for i, j, order in zip(bonds["i"], bonds["j"], bonds["bondorder"]):
                el1 = symbol[i]
                el2 = symbol[j]
                if order == 1:
                    txt.append(f"{el1}-{el2}")
                elif order == 2:
                    txt.append(f"{el1}={el2}")
                elif order == 3:
                    txt.append(f"{el1}#{el2}")
                else:
                    txt.append(f"{el1}-{order}-{el2}")
            tmp = tabulate(table, headers="keys", tablefmt="pretty")
            length = len(tmp.splitlines()[0])
            text += "\n"
            text += "Bonds".center(length)
            text += "\n"
            text += tmp
            text += "\n"

        # If requested, overwrite the bonding information in the system
        if P["use bonds"]:
            ids = configuration.atoms.ids
            if same:
                iatoms = [ids[i] for i in bonds["i"]]
                jatoms = [ids[j] for j in bonds["j"]]
                configuration.new_bondset()
                configuration.bonds.append(
                    i=iatoms, j=jatoms, bondorder=bonds["bondorder"]
                )
                text += "\nReplaced the bonds in the configuration with those from the "
                text += "Lewis structure.\n"
            else:
                iatoms = []
                jatoms = []
                bondorders = []
                for i in range(n_atoms):
                    for j in lneighbors[i]:
                        if j > i:
                            iatoms.append(ids[i])
                            jatoms.append(ids[j])
                            bondorders.append(1)
                configuration.new_bondset()
                configuration.bonds.append(
                    i=iatoms, j=jatoms, bondorder=bonds["bondorder"]
                )
                text += "\nReplaced the bonds in the configuration with those from the "
                text += "simple connectivity structure.\n"

        # Put any requested results into variables or tables
        self.store_results(
            configuration=configuration,
            data=data,
        )

        if "CPU_TIME" in data:
            t_total = data["CPU_TIME"]
            text += f"\nThe Lewis structure took a total of {t_total:.2f} s.\n"

        printer.normal(textwrap.indent(text, 4 * " "))



# -*- coding: utf-8 -*-

"""Top-level package for MOPAC step."""

# Bring up the classes so that they appear to be directly in
# the package.

from .mopac_base import MOPACBase  # noqa: F401

from .lewis_structure_step import LewisStructureStep  # noqa: F401
from .lewis_structure import LewisStructure  # noqa: F401
from .lewis_structure_parameters import LewisStructureParameters  # noqa: F401
from .tk_lewis_structure import TkLewisStructure  # noqa: F401

from .mopac_step import MOPACStep  # noqa: F401
from .mopac import MOPAC  # noqa: F401
from .tk_mopac import TkMOPAC  # noqa: F401
from .energy_step import EnergyStep  # noqa: F401
from .energy import Energy  # noqa: F401
from .energy_parameters import EnergyParameters  # noqa: F401
from .tk_energy import TkEnergy  # noqa: F401
from .forceconstants_step import ForceconstantsStep  # noqa: F401
from .forceconstants import Forceconstants  # noqa: F401
from .forceconstants_parameters import ForceconstantsParameters  # noqa: F401
from .tk_forceconstants import TkForceconstants  # noqa: F401
from .optimization_step import OptimizationStep  # noqa: F401
from .optimization import Optimization  # noqa: F401
from .optimization_parameters import OptimizationParameters  # noqa: F401
from .tk_optimization import TkOptimization  # noqa: F401
from .ir_step import IRStep  # noqa: F401
from .ir import IR  # noqa: F401
from .ir_parameters import IRParameters  # noqa: F401
from .tk_ir import TkIR  # noqa: F401
from .thermodynamics_step import ThermodynamicsStep  # noqa: F401
from .thermodynamics import Thermodynamics  # noqa: F401
from .thermodynamics_parameters import (  # noqa: F401
    ThermodynamicsParameters,
)
from .tk_thermodynamics import TkThermodynamics  # noqa: F401

from .metadata import metadata  # noqa: F401

# Handle versioneer
from ._version import get_versions

__author__ = """Paul Saxe"""
__email__ = "psaxe@molssi.org"
versions = get_versions()
__version__ = versions["version"]
__git_revision__ = versions["full-revisionid"]
del get_versions, versions



# -*- coding: utf-8 -*-
"""Control parameters for a MOPAC single-point energy calculation"""

import logging
import seamm

logger = logging.getLogger(__name__)


class LewisStructureParameters(seamm.Parameters):
    """The control parameters for creating a structure from SMILES"""

    parameters = {
        "atom cutoff": {
            "default": 30,
            "kind": "integer",
            "default_units": "",
            "enumeration": ("no printing", "unlimited"),
            "format_string": "s",
            "description": "No printing if more atoms than:",
            "help_text": "Limit for system size for printing",
        },
        "use bonds": {
            "default": "no",
            "kind": "boolean",
            "default_units": "",
            "enumeration": ("yes", "no"),
            "format_string": "s",
            "description": "Replace bonds in structure:",
            "help_text": "Replace the bonds on the structure.",
        },
        "adjust charge": {
            "default": "no",
            "kind": "boolean",
            "default_units": "",
            "enumeration": ("yes", "no"),
            "format_string": "s",
            "description": "Adjust charge on structure:",
            "help_text": "Adjust the charge on the structure.",
        },
        "ignore errors": {
            "default": "no",
            "kind": "boolean",
            "default_units": "",
            "enumeration": ("yes", "no"),
            "format_string": "s",
            "description": "Convert errors to warnings:",
            "help_text": "Only warn about errors, do not stop.",
        },
        "on errors use connectivity": {
            "default": "yes",
            "kind": "boolean",
            "default_units": "",
            "enumeration": ("yes", "no"),
            "format_string": "s",
            "description": "On errors, use connectivity instead:",
            "help_text": "Fallback to simple connectivity if Lewis structure fails.",
        },
        "results": {
            "default": {},
            "kind": "dictionary",
            "default_units": "",
            "enumeration": tuple(),
            "format_string": "",
            "description": "results",
            "help_text": "The results to save to variables or in tables.",
        },
    }

    def __init__(self, defaults={}, data=None):
        """Initialize the instance, by default from the default
        parameters given in the class"""

        super().__init__(
            defaults={**LewisStructureParameters.parameters, **defaults}, data=data
        )



# -*- coding: utf-8 -*-

"""The graphical part of a MOPAC Energy node"""

import logging
import tkinter as tk
import tkinter.ttk as ttk

import mopac_step
import seamm
import seamm_widgets as sw

logger = logging.getLogger(__name__)


class TkEnergy(seamm.TkNode):
    def __init__(
        self,
        tk_flowchart=None,
        node=None,
        canvas=None,
        x=120,
        y=20,
        w=200,
        h=50,
        my_logger=logger,
    ):
        """Initialize the graphical Tk MOPAC energy step

        Keyword arguments:
        """
        self.results_widgets = []

        super().__init__(
            tk_flowchart=tk_flowchart,
            node=node,
            canvas=canvas,
            x=x,
            y=y,
            w=w,
            h=h,
            my_logger=my_logger,
        )

    def right_click(self, event):
        """Probably need to add our dialog..."""

        super().right_click(event)
        self.popup_menu.add_command(label="Edit..", command=self.edit)

        self.popup_menu.tk_popup(event.x_root, event.y_root, 0)

    def create_dialog(self, title="Edit MOPAC Energy Step"):
        """Create the dialog!"""
        self.logger.debug("Creating the dialog")
        frame = super().create_dialog(title=title, widget="notebook", results_tab=True)

        P = self.node.parameters

        # Just write input
        self["input only"] = P["input only"].widget(frame)

        # Frame to isolate widgets
        e_frame = self["energy frame"] = ttk.LabelFrame(
            frame,
            borderwidth=4,
            relief="sunken",
            text="Hamiltonian Parameters",
            labelanchor="n",
            padding=10,
        )

        # Create all the widgets
        for key in mopac_step.EnergyParameters.parameters:
            if key not in ("results", "extra keywords", "create tables", "input only"):
                self[key] = P[key].widget(e_frame)

        # Set the callbacks for changes
        for widget in ("calculation", "convergence", "MOZYME", "COSMO"):
            w = self[widget]
            w.combobox.bind("<<ComboboxSelected>>", self.reset_energy_frame)
            w.combobox.bind("<Return>", self.reset_energy_frame)
            w.combobox.bind("<FocusOut>", self.reset_energy_frame)

        self.setup_results()

        self.logger.debug("Finished creating the dialog")

        return frame

    def reset_dialog(self, widget=None):
        frame = self["frame"]
        for slave in frame.grid_slaves():
            slave.grid_forget()

        row = 0
        # Whether to just write input
        self["input only"].grid(row=row, column=0, sticky=tk.W)
        row += 1

        # Put in the energy frame
        self["energy frame"].grid(row=row, column=0, sticky=tk.EW)
        row += 1

        # and the widgets in it
        self.reset_energy_frame()

        return row

    def reset_energy_frame(self, widget=None):
        frame = self["energy frame"]
        for slave in frame.grid_slaves():
            slave.grid_forget()

        calculation = self["calculation"].get()
        convergence = self["convergence"].get()
        cosmo = self["COSMO"].get()
        mozyme = self["MOZYME"].get()

        widgets = []
        row = 0
        for key in ("hamiltonian", "calculation"):
            self[key].grid(row=row, column=0, columnspan=2, sticky=tk.EW)
            widgets.append(self[key])
            row += 1

        if "hf" in calculation.lower():
            for key in ("uhf",):
                self[key].grid(row=row, column=0, columnspan=2, sticky=tk.EW)
                widgets.append(self[key])
                row += 1

        if "ci" in calculation.lower():
            for key in (
                "number ci orbitals",
                "number doubly occupied ci orbitals",
                "ci root",
                "print ci details",
            ):
                self[key].grid(row=row, column=0, columnspan=2, sticky=tk.EW)
                widgets.append(self[key])
                row += 1

        for key in ("convergence",):
            self[key].grid(row=row, column=0, columnspan=2, sticky=tk.EW)
            widgets.append(self[key])
            row += 1

        if convergence == "relative":
            self["relative"].grid(row=row, column=1, sticky=tk.W)
            row += 1
        elif convergence == "absolute":
            self["absolute"].grid(row=row, column=1, sticky=tk.W)
            row += 1
        elif convergence not in ("normal", "precise"):
            # variable ... so put in all possibilities
            self["relative"].grid(row=row, column=1, sticky=tk.W)
            row += 1
            self["absolute"].grid(row=row, column=1, sticky=tk.W)
            row += 1
            sw.align_labels((self["relative"], self["absolute"]), sticky=tk.E)

        if "hf" in calculation.lower():
            self["MOZYME"].grid(row=row, column=0, columnspan=2, sticky=tk.EW)
            widgets.append(self["MOZYME"])
            row += 1
            subwidgets = []
            if mozyme != "always" and mozyme != "never":
                self["nMOZYME"].grid(row=row, column=1, sticky=tk.W)
                subwidgets.append(self["nMOZYME"])
                row += 1
            if mozyme != "never":
                self["MOZYME follow-up"].grid(row=row, column=1, sticky=tk.W)
                subwidgets.append(self["MOZYME follow-up"])
                row += 1
            sw.align_labels(subwidgets, sticky=tk.E)

        self["COSMO"].grid(row=row, column=0, columnspan=2, sticky=tk.EW)
        widgets.append(self["COSMO"])
        row += 1
        if cosmo == "yes":
            widgets1 = []

            for key in ("eps", "rsolve", "nspa", "disex"):
                self[key].grid(row=row, column=1, sticky=tk.W)
                widgets1.append(self[key])
                row += 1
            sw.align_labels(widgets1, sticky=tk.E)

        for key in ("calculate gradients", "bond orders"):
            self[key].grid(row=row, column=0, columnspan=2, sticky=tk.EW)
            widgets.append(self[key])
            row += 1

        sw.align_labels(widgets, sticky=tk.E)
        frame.columnconfigure(0, minsize=100)

        return row



# -*- coding: utf-8 -*-

"""The graphical part of the Lewsi structure node"""

import logging
import tkinter as tk

import mopac_step
import seamm
import seamm_widgets as sw

logger = logging.getLogger(__name__)


class TkLewisStructure(seamm.TkNode):
    def __init__(
        self,
        tk_flowchart=None,
        node=None,
        canvas=None,
        x=120,
        y=20,
        w=200,
        h=50,
        my_logger=logger,
    ):
        """Initialize the graphical Tk MOPAC energy step

        Keyword arguments:
        """
        self.results_widgets = []

        super().__init__(
            tk_flowchart=tk_flowchart,
            node=node,
            canvas=canvas,
            x=x,
            y=y,
            w=w,
            h=h,
            my_logger=my_logger,
        )

    def right_click(self, event):
        """Probably need to add our dialog..."""

        super().right_click(event)
        self.popup_menu.add_command(label="Edit..", command=self.edit)

        self.popup_menu.tk_popup(event.x_root, event.y_root, 0)

    def create_dialog(self, title="Edit Lewis Structure Step"):
        """Create the dialog!"""
        self.logger.debug("Creating the dialog")
        frame = super().create_dialog(title=title)

        # Create all the widgets
        P = self.node.parameters
        row = 0
        widgets = []
        for key in mopac_step.LewisStructureParameters.parameters:
            if key not in ("results",):
                w = self[key] = P[key].widget(frame)
                w.grid(row=row, column=0, sticky=tk.EW)
                widgets.append(w)
                row += 1
        sw.align_labels(widgets, sticky=tk.E)

        self.setup_results()

        self.logger.debug("Finished creating the dialog")

        return frame



# -*- coding: utf-8 -*-

"""Setup and run MOPAC"""

import logging
import re

import seamm
import seamm_util.printing as printing
import mopac_step

logger = logging.getLogger(__name__)
job = printing.getPrinter()
printer = printing.getPrinter("mopac")


class MOPACBase(seamm.Node):
    def __init__(
        self,
        flowchart=None,
        namespace=None,
        extension=None,
        title=None,
        logger=logger,
    ):
        """Initialize the node"""

        logger.debug("Creating MOPACBase {}".format(self))

        super().__init__(
            flowchart=flowchart, title=title, extension=extension, logger=logger
        )

    @property
    def version(self):
        """The semantic version of this module."""
        return mopac_step.__version__

    @property
    def git_revision(self):
        """The git version of this module."""
        return mopac_step.__git_revision__

    def create_parser(self):
        """Setup the command-line / config file parser"""
        parser_name = self.step_type
        parser = self.flowchart.parser

        # Remember if the parser exists ... this type of step may have been
        # found before
        parser_exists = parser.exists(parser_name)

        # Create the standard options, e.g. log-level
        result = super().create_parser(name=parser_name)

        if parser_exists:
            return result

        # Options for Mopac
        parser.add_argument(
            parser_name,
            "--mopac-exe",
            default="mopac",
            help="the name of the MOPAC executable",
        )

        parser.add_argument(
            parser_name,
            "--mopac-path",
            default="",
            help="the path to the MOPAC executable",
        )

        parser.add_argument(
            parser_name,
            "--ncores",
            default="default",
            help="How many threads to use in MOPAC",
        )

        parser.add_argument(
            parser_name,
            "--mkl-num-threads",
            default="default",
            help="How many threads to use with MKL in MOPAC",
        )

        parser.add_argument(
            parser_name,
            "--max-atoms-to-print",
            default=25,
            help="Maximum number of atoms to print charges, etc.",
        )

        return result

    def mopac_structure(self):
        """Create the input for the structure."""
        _, configuration = self.get_system_configuration(None)

        structure = ""
        atoms = configuration.atoms
        elements = atoms.symbols
        coordinates = atoms.get_coordinates(fractionals=False)
        if "freeze" in atoms:
            freeze = atoms["freeze"]
        else:
            freeze = [""] * len(elements)
        for element, xyz, frz in zip(elements, coordinates, freeze):
            x, y, z = xyz
            line = "{:2} {: 12.8f} {:d} {: 12.8f} {:d} {: 12.8f} {:d}\n".format(
                element,
                x,
                0 if "x" in frz else 1,
                y,
                0 if "y" in frz else 1,
                z,
                0 if "z" in frz else 1,
            )
            structure += line

        symlines = ""
        if configuration.periodicity == 3:
            # The three translation vectors
            element = "Tv"
            uvw = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
            XYZ = configuration.cell.to_cartesians(uvw)
            n = configuration.n_atoms
            tv1 = n + 1
            tv2 = n + 2
            tv3 = n + 3
            if self._lattice_opt:
                if self._lattice_shear:
                    freeze = [[1, 1, 1], [1, 1, 1], [1, 1, 1]]
                else:
                    freeze = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
                if self._lattice_couple == "x, y and z":
                    freeze[1][1] = 0
                    freeze[2][2] = 0
                    symlines += f"    {tv1} 14 {tv2}\n"
                    symlines += f"    {tv1}  9 {tv3}\n"
                elif self._lattice_couple == "x and y":
                    freeze[1][1] = 0
                    symlines += f"    {tv1} 14 {tv2}\n"
                elif self._lattice_couple == "x and z":
                    freeze[2][2] = 0
                    symlines += f"    {tv1} 9 {tv2}\n"
                elif self._lattice_couple == "y and z":
                    freeze[2][2] = 0
                    symlines += f"    {tv2} 15 {tv3}\n"
            else:
                freeze = [[0, 0, 0], [0, 0, 0], [0, 0, 0]]
            for xyz, frz in zip(XYZ, freeze):
                x, y, z = xyz
                fx, fy, fz = frz
                line = f"{element:2} {x: 12.8f} {fx} {y: 12.8f} {fy} {z: 12.8f} {fz}\n"
                structure += line

        return structure, symlines

    def parse_arc(self, filename="mopac.arc"):
        """Digest the ARC file and get the coordinates.

        Parameters
        ----------
        filename : str
            The name of the ARC file

        Returns
        -------
        coordinates : [n_atoms*[3]]
        """
        xyz = None
        cell_vectors = []
        with open(filename, "r") as fd:
            for line in fd:
                if "FINAL GEOMETRY OBTAINED" in line:
                    xyz = []
                    ii = 0
                    for line in fd:
                        if " &" not in line and " +" not in line and line[0] != "*":
                            ii += 1
                            if ii == 3:
                                break
                    for line in fd:
                        line = line.strip()
                        if line == "":
                            break
                        if line[0] != "*":
                            symbol, x, fx, y, fy, z, fz = line.split()
                            if symbol == "Tv":
                                cell_vectors.append(
                                    [float(x), float(y), float(z)]
                                )  # yapf: disable
                            else:
                                xyz.append([float(x), float(y), float(z)])
        return xyz, cell_vectors

    def parse_aux(self, lines):
        """Digest a section of the aux file"""

        properties = mopac_step.metadata["results"]
        trans = str.maketrans("Dd", "Ee")

        data = {}
        lineno = -1
        nlines = len(lines)

        spin_polarized = False

        while True:
            lineno += 1
            if lineno >= nlines:
                break
            line = lines[lineno].strip()
            if "END OF MOPAC PROGRAM" in line:
                continue
            if "END OF MOPAC FILE" in line:
                continue
            if line[0] == "#":
                continue
            if "=" not in line:
                raise RuntimeError("Problem parsing MOPAC aux file: '" + line + "'")
            key, rest = line.split("=", maxsplit=1)
            if key[-1] == "]":
                name, size = key[0:-1].split("[")
                size = int(size.lstrip("0"))
                if ":" in name:
                    name, units = name.split(":")

                if name not in properties:
                    logger.warning("Property '{}' not recognized.".format(name))
                    kind = "string"
                else:
                    kind = properties[name]["type"]
                    if "units" in properties[name]:
                        data[name + ",units"] = properties[name]["units"]

                if name == "NUM_ALPHA_ELECTRONS":
                    spin_polarized = True

                # Bug workaround
                # Sometimes MOPAC does not write out the MO occupancies
                if name == "MOLECULAR_ORBITAL_OCCUPANCIES":
                    tmp_line = lines[lineno + 1].strip()
                    if tmp_line[0].isalpha():
                        continue
                # end of workaround

                # Check for floating point numbers run together
                if kind == "float":
                    values = []
                    for value in rest.split():
                        tmp = value.split(".")
                        if len(tmp) <= 2:
                            values.append(value)
                        else:
                            # Run together ... lets see how many decimals
                            n_decimals = len(tmp[-1])
                            # and before the decimal
                            n_digits = len(tmp[-2]) - n_decimals
                            n = n_digits + 1 + n_decimals
                            n_values = len(tmp) - 1
                            # blanks at front have been stripped, so count back
                            start = 0
                            end = len(value) - (n_values - 1) * n
                            while start < len(value):
                                values.append(value[start:end])
                                start = end
                                end += n
                    tmp = values
                else:
                    tmp = rest.split()

                # AUX file contains alpha and beta orbitals, that's why we are
                # doubling it
                if name == "MICROSTATE_CONFIGURATIONS" and spin_polarized:
                    size = size * 2

                while len(tmp) < size:
                    lineno += 1
                    line = lines[lineno].strip()
                    if line[0] != "#":
                        tmp.extend(line.split())

                if kind == "integer":
                    values = []
                    for value in tmp:
                        values.append(int(value))
                elif kind == "float":
                    values = []
                    for value in tmp:
                        values.append(float(value.translate(trans)))
                else:
                    values = tmp

                if "UPDATED" in name:
                    if name not in data:
                        data[name] = []
                    data[name].append(values)
                else:
                    if (
                        name in properties
                        and properties[name]["dimensionality"] == "scalar"
                    ):
                        if not (units == "ARBITRARY_UNITS" and name in data):
                            data[name] = values[0]
                    else:
                        data[name] = values
            else:
                if ":" in key:
                    name, units = key.split(":")
                else:
                    name = key

                if name not in properties:
                    logger.warning("Property '{}' not recognized.".format(name))
                    kind = "string"
                else:
                    kind = properties[name]["type"]
                if "units" in properties[name]:
                    data[name + ",units"] = properties[name]["units"]
                if kind == "integer":
                    value = int(rest)
                elif kind == "float":
                    value = float(self._sanitize_value(rest.strip()))
                else:
                    value = rest.strip('"')

                if "UPDATED" in name:
                    if name not in data:
                        data[name] = []
                    data[name].append(value)
                else:
                    data[name] = value

        return data

    def _sanitize_value(self, value):
        regex = r"^([-+]?[.0-9]+)([EeDd]*)([-+][0-9]+)$"
        subs = r"\1E\3"
        ret = float(re.sub(regex, subs, value))
        return ret



# -*- coding: utf-8 -*-

"""Setup and run MOPAC"""

import logging
from pathlib import Path
import textwrap
import traceback

import numpy as np
from tabulate import tabulate

import mopac_step
from molsystem import RMSD
import seamm
import seamm_util.printing as printing
from seamm_util import units_class
from seamm_util.printing import FormattedText as __

logger = logging.getLogger(__name__)
job = printing.getPrinter()
printer = printing.getPrinter("mopac")


class Optimization(mopac_step.Energy):
    def __init__(self, flowchart=None, title="Optimization", extension=None):
        """Initialize the node"""

        logger.debug("Creating Optimization {}".format(self))

        super().__init__(flowchart=flowchart, title=title, extension=extension)

        self._calculation = "optimization"
        self._model = None
        self._metadata = mopac_step.metadata
        self.parameters = mopac_step.OptimizationParameters()
        self.description = "A structural optimization"

    def description_text(self, P=None):
        """Prepare information about what this node will do"""

        if not P:
            P = self.parameters.values_to_dict()

        # The energy part of the description
        tmp = super().description_text(P)
        energy_description = textwrap.dedent("\n".join(tmp.splitlines()[1:]))

        # Hamiltonian followed by convergence
        text = "Geometry optimization with the "
        if P["method"] == "default":
            text += "default optimizer (EF for small systems," " L-BFGS for larger)."
        elif P["method"][0:1] == "EF":
            text += "eigenvector following (EF) method."
        elif P["method"][0:3] == "BFGS":
            text += "BFGS method."
        elif P["method"][0:5] == "L-BFGS":
            text += "L-BFGS small memory version of the BFGS method."
        elif P["method"].startswith("TS"):
            text += "EF method for a transition state."
        elif P["method"].startswith("SIGMA"):
            text += "McIver-Komornicki method for transition states."
        elif P["method"].startswith("NLLSQ"):
            text += "Bartel's method of nonlinear least squares of the gradient."
        else:
            text += "optimization method determined at runtime by '{method}'."

        if P["LatticeOpt"]:
            pressure = P["pressure"]
            if isinstance(pressure, units_class):
                pressure = f"{pressure:~P}"
            text += (
                " The cell for periodic systems will also be optimized, with an "
                f"applied pressure of {pressure}."
            )
        else:
            text += " The cell for periodic systems will not be optimized."

        if P["gnorm"] == "default":
            text += " The geometrical convergence is the default of 1.0 kcal/mol/Å."
        elif P["gnorm"][0] == "$":
            text += (
                " The geometrical convergence will be determined "
                "at runtime by '{gnorm}'."
            )
        else:
            text += " The geometrical convergence is {gnorm}."

        # Put in the description of the energy calculation
        text += "\n\n" + energy_description
        text += "\n\n"

        if self.is_expr(P["hamiltonian"]):
            kwargs = {"Hamiltonian": "{{" + P["hamiltonian"] + "}}"}
        else:
            kwargs = {"Hamiltonian": P["hamiltonian"]}
        text += seamm.standard_parameters.structure_handling_description(P, **kwargs)

        return self.header + "\n" + __(text, **P, indent=4 * " ").__str__()

    def get_input(self):
        """Get the input for an optimization MOPAC"""

        references = self.parent.references
        bibliography = self.parent._bibliography

        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )

        # Have to fix formatting for printing...
        PP = dict(P)
        for key in PP:
            if isinstance(PP[key], units_class):
                PP[key] = "{:~P}".format(PP[key])

        system, configuration = self.get_system_configuration(None)

        # Let parent know about cell optimization
        self.parent._lattice_opt = P["LatticeOpt"]
        self.parent._lattice_shear = P["allow shear"]
        self.parent._lattice_couple = P["couple"]

        # Get the inputs from the energy class. This also sets the description properly.
        inputs = super().get_input()

        # Remove the 1SCF keyword from the energy setup
        # 'keywords' is a reference, so will change in situ.
        keywords, _, _ = inputs[0]
        if "1SCF" in keywords:
            keywords.remove("1SCF")

        # Pressure
        if configuration.periodicity == 3 and P["LatticeOpt"]:
            pressure = P["pressure"]
            pressure = pressure.to("GPa").magnitude
            keywords.append(f"P={pressure}GPa")

        # and the optimization-specific parts
        method = P["method"]
        if method == "default":
            pass
        elif method[0:2] == "EF":
            keywords.append("EF")
            if P["recalc"] != "never":
                keywords.append(f'RECALC={P["recalc"]}')
            if str(P["dmax"]) != self.parameters["dmax"].default:
                keywords.append("DMAX={}".format(P["dmax"]))
            references.cite(
                raw=bibliography["Baker_1986"],
                alias="Baker_1986",
                module="mopac_step",
                level=1,
                note="Eigenvector-following minimizer.",
            )
        elif method[0:4] == "BFGS":
            keywords.append("BFGS")
            references.cite(
                raw=bibliography["Broyden_1970"],
                alias="Broyden_1970",
                module="mopac_step",
                level=1,
                note="Broyden-Fletcher-Goldfarb-Shanno (BFGS) minimizer.",
            )
            references.cite(
                raw=bibliography["Fletcher_1970"],
                alias="Fletcher_1970",
                module="mopac_step",
                level=1,
                note="Broyden-Fletcher-Goldfarb-Shanno (BFGS) minimizer.",
            )
            references.cite(
                raw=bibliography["Goldfarb_1970"],
                alias="Goldfarb_1970",
                module="mopac_step",
                level=1,
                note="Broyden-Fletcher-Goldfarb-Shanno (BFGS) minimizer.",
            )
            references.cite(
                raw=bibliography["Shanno_1970"],
                alias="Shanno_1970",
                module="mopac_step",
                level=1,
                note="Broyden-Fletcher-Goldfarb-Shanno (BFGS) minimizer.",
            )
            references.cite(
                raw=bibliography["Thiel_1988"],
                alias="Thiel_1988",
                module="mopac_step",
                level=1,
                note="Broyden-Fletcher-Goldfarb-Shanno (BFGS) minimizer.",
            )
        elif method[0:6] == "L-BFGS":
            keywords.append("LBFGS")
            references.cite(
                raw=bibliography["Nocedal_1980"],
                alias="Nocedal_1980",
                module="mopac_step",
                level=1,
                note="Limited-memory BFGS (L-BFGS) minimizer.",
            )
        elif method[0:2] == "TS":
            keywords.append("TS")
            if P["recalc"] != "never":
                keywords.append(f'RECALC={P["recalc"]}')
            if str(P["dmax"]) != self.parameters["dmax"].default:
                keywords.append("DMAX={}".format(P["dmax"]))
            references.cite(
                raw=bibliography["Baker_1986"],
                alias="Baker_1986",
                module="mopac_step",
                level=1,
                note="Eigenvector-following minimizer for transition states.",
            )
        elif method[0:5] == "SIGMA":
            keywords.append("SIGMA")
            references.cite(
                raw=bibliography["McIver_1971"],
                alias="McIver_1971",
                module="mopac_step",
                level=1,
                note="Gradient-norm minimizer for transition states.",
            )
            references.cite(
                raw=bibliography["McIver_1972"],
                alias="McIver_1972",
                module="mopac_step",
                level=1,
                note="Gradient-norm minimizer for transition states.",
            )
        elif method[0:5] == "NLLSQ":
            keywords.append("NLLSQ")
            references.cite(
                raw=bibliography["Bartels_1972"],
                alias="Bartels_1972",
                module="mopac_step",
                level=1,
                note="NLLSQ gradient-norm minimizer for transition states.",
            )
        else:
            text = "Don't recognize optimization method '{}'".format(P["method"])
            logger.critical(text)
            raise RuntimeError(text)

        if P["cycles"] != "unlimited":
            keywords.append("CYCLES={}".format(P["cycles"]))
        if P["convergence"] not in ("normal", "precise"):
            if P["gnorm"] != self.parameters["gnorm"].default:
                keywords.append("GNORM={}".format(P["gnorm"]))

        return inputs

    def analyze(self, indent="", data_sections=[], out_sections=[], table=None):
        """Parse the output and generating the text output and store the
        data in variables for other stages to access
        """

        # Get the parameters used
        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )

        initial_system, initial_configuration = self.get_system_configuration(None)

        # Get the data.
        data = data_sections[0]

        # If the optimizer used was the default, put in the correct citations

        references = self.parent.references
        bibliography = self.parent._bibliography

        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )

        if P["method"] == "default":
            tmp = "\n".join(out_sections[0])
            if (
                "GEOMETRY OPTIMISED USING EIGENVECTOR FOLLOWING (EF)" in tmp
                or "Geometry optimization using EF" in tmp
            ):
                opt_method = "EF -- eigenvector following"
                references.cite(
                    raw=bibliography["Baker_1986"],
                    alias="Baker_1986",
                    module="mopac_step",
                    level=1,
                    note="Eigenvector-following minimizer.",
                )
            elif (
                "Geometry optimization using BFGS" in tmp or "SATISFIED IN BFGS" in tmp
            ):
                opt_method = "BFGS -- Broyden-Fletcher-Goldfarb-Shanno algorithm"
                references.cite(
                    raw=bibliography["Broyden_1970"],
                    alias="Broyden_1970",
                    module="mopac_step",
                    level=1,
                    note="Broyden-Fletcher-Goldfarb-Shanno (BFGS) minimizer.",
                )
                references.cite(
                    raw=bibliography["Fletcher_1970"],
                    alias="Fletcher_1970",
                    module="mopac_step",
                    level=1,
                    note="Broyden-Fletcher-Goldfarb-Shanno (BFGS) minimizer.",
                )
                references.cite(
                    raw=bibliography["Goldfarb_1970"],
                    alias="Goldfarb_1970",
                    module="mopac_step",
                    level=1,
                    note="Broyden-Fletcher-Goldfarb-Shanno (BFGS) minimizer.",
                )
                references.cite(
                    raw=bibliography["Shanno_1970"],
                    alias="Shanno_1970",
                    module="mopac_step",
                    level=1,
                    note="Broyden-Fletcher-Goldfarb-Shanno (BFGS) minimizer.",
                )
                references.cite(
                    raw=bibliography["Thiel_1988"],
                    alias="Thiel_1988",
                    module="mopac_step",
                    level=1,
                    note="Broyden-Fletcher-Goldfarb-Shanno (BFGS) minimizer.",
                )
            elif (
                "Geometry optimization using L-BFGS" in tmp
                or "SATISFIED IN L-BFGS" in tmp
            ):
                opt_method = (
                    "L-BFGS -- Limited memory Broyden-Fletcher-Goldfarb-Shanno "
                    "algorithm"
                )
                references.cite(
                    raw=bibliography["Broyden_1970"],
                    alias="Broyden_1970",
                    module="mopac_step",
                    level=1,
                    note="Broyden-Fletcher-Goldfarb-Shanno (BFGS) minimizer.",
                )
                references.cite(
                    raw=bibliography["Fletcher_1970"],
                    alias="Fletcher_1970",
                    module="mopac_step",
                    level=1,
                    note="Broyden-Fletcher-Goldfarb-Shanno (BFGS) minimizer.",
                )
                references.cite(
                    raw=bibliography["Goldfarb_1970"],
                    alias="Goldfarb_1970",
                    module="mopac_step",
                    level=1,
                    note="Broyden-Fletcher-Goldfarb-Shanno (BFGS) minimizer.",
                )
                references.cite(
                    raw=bibliography["Shanno_1970"],
                    alias="Shanno_1970",
                    module="mopac_step",
                    level=1,
                    note="Broyden-Fletcher-Goldfarb-Shanno (BFGS) minimizer.",
                )
                references.cite(
                    raw=bibliography["Thiel_1988"],
                    alias="Thiel_1988",
                    module="mopac_step",
                    level=1,
                    note="Broyden-Fletcher-Goldfarb-Shanno (BFGS) minimizer.",
                )
            else:
                logger.warning("Could not find which minimizer was used!")
                opt_method = "default -- but could not find which was used"
        else:
            opt_method = P["method"]

        # The results
        if "NUMBER_SCF_CYCLES" in data_sections[0]:
            text = (
                f"The geometry optimization using {opt_method} -- converged in "
                "{NUMBER_SCF_CYCLES} iterations."
            )
        else:
            data["NUMBER_SCF_CYCLES"] = len(data["HEAT_OF_FORM_UPDATED"])
            data["HEAT_OF_FORMATION"] = data["HEAT_OF_FORM_UPDATED"][-1]
            data["GRADIENT_NORM"] = data["GRADIENT_NORM_UPDATED"][-1]
            text = (
                f"The geometry optimization using {opt_method} -- did not converge in "
                "{NUMBER_SCF_CYCLES} steps. The following results are for the final "
                "structure.\n"
            )

        # Update the structure
        periodicity = initial_configuration.periodicity
        if "ATOM_X_OPT" in data or "ATOM_X_UPDATED" in data:
            initial_RDKMol = initial_configuration.to_RDKMol()

            if P["structure handling"] == "Discard the structure":
                system = initial_configuration
                configuration = initial_configuration
            else:
                system, configuration = self.get_system_configuration(
                    P, same_as=initial_configuration
                )
                if periodicity != 0 and P["LatticeOpt"]:
                    if "TRANS_VECTS" in data:
                        vectors = data["TRANS_VECTS"]
                        lattice = [
                            [vectors[0], vectors[1], vectors[2]],
                            [vectors[3], vectors[4], vectors[5]],
                            [vectors[6], vectors[7], vectors[8]],
                        ]
                        configuration.cell.from_vectors(lattice)
                    elif "TRANS_VECTS_UPDATED" in data:
                        vectors = data["TRANS_VECTS_UPDATED"][-1]
                        lattice = [
                            [vectors[0], vectors[1], vectors[2]],
                            [vectors[3], vectors[4], vectors[5]],
                            [vectors[6], vectors[7], vectors[8]],
                        ]
                        configuration.cell.from_vectors(lattice)
                    else:
                        logger.warning(
                            "Expected updated lattice vectors, but did not find!"
                        )
            xyz = []
            if "ATOM_X_OPT" in data:
                it = iter(data["ATOM_X_OPT"])
            else:
                it = iter(data["ATOM_X_UPDATED"][-1])
            for x in it:
                xyz.append([float(x), float(next(it)), float(next(it))])

            if configuration.symmetry.n_symops > 1:
                # Convert to coordinates of just the asymmetric atoms.
                xyz, delta = configuration.symmetry.symmetrize_coordinates(
                    xyz, fractionals=False
                )
                delta = np.array(delta)
                displacement = np.linalg.norm(delta, axis=1)
                max_disp = displacement.max()
                text += (
                    f"\nThe largest displacement from symmetry was {max_disp:.6f} Å.\n"
                )

            if P["structure handling"] == "Discard the structure":
                RDKMol = initial_configuration.to_RDKMol()
                RDKMol.GetConformer(0).SetPositions(np.array(xyz))
            else:
                configuration.atoms.set_coordinates(xyz, fractionals=False)
                RDKMol = configuration.to_RDKMol()

            result = RMSD(RDKMol, initial_RDKMol, symmetry=True, include_h=True)
            data["RMSD with H"] = result["RMSD"]
            data["displaced atom with H"] = result["displaced atom"]
            data["maximum displacement with H"] = result["maximum displacement"]

            # Align the structure
            if P["structure handling"] != "Discard the structure":
                configuration.coordinates_from_RDKMol(RDKMol)

            result = RMSD(RDKMol, initial_RDKMol, symmetry=True)
            data["RMSD"] = result["RMSD"]
            data["displaced atom"] = result["displaced atom"]
            data["maximum displacement"] = result["maximum displacement"]

            text += seamm.standard_parameters.set_names(
                system, configuration, P, _first=True, Hamiltonian=P["hamiltonian"]
            )

        # Write the structure out for viewing.
        directory = Path(self.directory)
        directory.mkdir(parents=True, exist_ok=True)

        if periodicity == 0:
            configuration.to_sdf(directory / "optimized.sdf")
        else:
            #  MMCIF file has bonds
            try:
                path = directory / "optimized.mmcif"
                path.write_text(configuration.to_mmcif_text())
            except Exception:
                message = "Error creating the mmcif file\n\n" + traceback.format_exc()
                logger.warning(message)
            # CIF file has cell
            if configuration.periodicity == 3:
                try:
                    path = directory / "optimized.cif"
                    path.write_text(configuration.to_cif_text())
                except Exception:
                    message = "Error creating the cif file\n\n" + traceback.format_exc()
                    logger.warning(message)

        printer.normal(__(text, **data, indent=8 * " "))

        super().analyze(
            indent=indent,
            data_sections=data_sections,
            out_sections=out_sections,
            table=table,
        )

        # Tabulate the change in cell parameters
        # For periodic systems, the change in cell
        if initial_configuration.periodicity != 0:
            a0, b0, c0, alpha0, beta0, gamma0 = initial_configuration.cell.parameters
            a, b, c, alpha, beta, gamma = configuration.cell.parameters
            data["a"] = a
            data["b"] = b
            data["c"] = c
            data["alpha"] = alpha
            data["beta"] = beta
            data["gamma"] = gamma
            data["delta a"] = a - a0
            data["delta b"] = b - b0
            data["delta c"] = c - c0
            data["delta alpha"] = alpha - alpha0
            data["delta beta"] = beta - beta0
            data["delta gamma"] = gamma - gamma0

            V0 = initial_configuration.cell.volume
            V = configuration.cell.volume
            rho0 = initial_configuration.density
            rho = configuration.density
            data["V"] = V
            data["density"] = rho
            data["delta V"] = V - V0
            data["delta density"] = rho - rho0

            ctable = {
                "": ("𝗮", "𝗯", "𝗰", "𝞪", "𝞫", "𝞬", "V", "ρ"),
                "Initial": (
                    f"{a0:.3f}",
                    f"{b0:.3f}",
                    f"{c0:.3f}",
                    f"{alpha0:.1f}",
                    f"{beta0:.1f}",
                    f"{gamma0:.1f}",
                    f"{V0:.1f}",
                    f"{rho0:.3f}",
                ),
                "Final": (
                    f"{a:.3f}",
                    f"{b:.3f}",
                    f"{c:.3f}",
                    f"{alpha:.1f}",
                    f"{beta:.1f}",
                    f"{gamma:.1f}",
                    f"{V:.1f}",
                    f"{rho:.3f}",
                ),
                "Change": (
                    f"{a - a0:.3f}",
                    f"{b - b0:.3f}",
                    f"{c - c0:.3f}",
                    f"{alpha - alpha0:.1f}",
                    f"{beta - beta0:.1f}",
                    f"{gamma - gamma0:.1f}",
                    f"{V - V0:.1f}",
                    f"{rho - rho0:.3f}",
                ),
                "Units": ("Å", "Å", "Å", "°", "°", "°", "Å³", "g/mL"),
            }

            tmp = tabulate(
                ctable,
                headers="keys",
                tablefmt="rounded_outline",
                colalign=("center", "decimal", "decimal", "decimal", "center"),
                disable_numparse=True,
            )
            length = len(tmp.splitlines()[0])
            text_lines = []
            text_lines.append("Cell Parameters".center(length))
            text_lines.append(tmp)
            printer.normal(
                textwrap.indent("\n".join(text_lines), self.indent + 7 * " ")
            )
            printer.normal("")



# -*- coding: utf-8 -*-

"""Setup and run MOPAC"""

import copy
import csv
import logging
from math import sqrt
from pathlib import Path
import pprint  # noqa: F401
import textwrap

import numpy as np
from tabulate import tabulate

import mopac_step
import seamm
import seamm.data
from seamm_util import Q_, units_class
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __

logger = logging.getLogger(__name__)
job = printing.getPrinter()
printer = printing.getPrinter("mopac")


class Energy(seamm.Node):
    def __init__(self, flowchart=None, title="Energy", extension=None):
        """Initialize the node"""

        logger.debug("Creating Energy {}".format(self))

        super().__init__(flowchart=flowchart, title=title, extension=extension)

        self._calculation = "energy"
        self._model = None
        self._metadata = mopac_step.metadata
        self._use_mozyme = None
        self.parameters = mopac_step.EnergyParameters()
        self.description = "A single point energy calculation"

    @property
    def header(self):
        """A printable header for this section of output"""
        return "Step {}: {}".format(".".join(str(e) for e in self._id), self.title)

    @property
    def version(self):
        """The semantic version of this module."""
        return mopac_step.__version__

    @property
    def git_revision(self):
        """The git version of this module."""
        return mopac_step.__git_revision__

    def description_text(self, P=None):
        """Prepare information about what this node will do"""

        if not P:
            P = self.parameters.values_to_dict()

        calculation = P["calculation"].lower()
        if self.is_expr(calculation):
            text = "The type of calculation will be determined by {calculation}, which "
        elif "hf" in calculation:
            text = "The Hartree-Fock calculation "
        elif "ci" in calculation:
            if "cisdt" in calculation:
                text = "This will be a CI calculation with single, double, and triple "
                text += "excitations"
            elif "cisd" in calculation:
                text = "This will be a CI calculation with single and double "
                text += "excitations"
            elif "cis" in calculation:
                text = "This will be a CI calculation with single excitations"
            elif "casci" in calculation:
                text = "This will be a complete active space (CAS) CI "
            n_docc = P["number doubly occupied ci orbitals"]
            if n_docc == "default":
                text += " using {number ci orbitals} orbitals centered on the highest "
                text += "doubly occupied orbital."
            else:
                text += " using {number ci orbitals} with "
                text += "{number doubly occupied ci orbitals} doubly occupied "
                text += "orbitals."
            text += " If there are an odd number of electrons, the singly occupied "
            text += "orbital will be in the calculation, so there will be one less "
            text += "virtual orbital than in the closed-shell case."
            text += " The CI calculations will use orbitals from a Hartree-Fock "
            text += "calculation that "
        text += "will use the Hamiltonian {hamiltonian}, converged to "
        # Convergence
        if P["convergence"] == "normal":
            text += "the 'normal' level of 1.0e-04 kcal/mol."
        elif P["convergence"] == "precise":
            text += "the 'precise' level of 1.0e-06 kcal/mol."
        elif P["convergence"] == "relative":
            text += "a factor of {relative} times the normal criterion."
        elif P["convergence"] == "absolute":
            text += "converged to {absolute}."

        if "hf" in calculation:
            if self.parameters["uhf"].is_expr:
                text += (
                    " Whether to use spin-unrestricted SCF (UHF) for closed-shell "
                    "molecules will be determined by '{uhf}'."
                )
            elif self.parameters["uhf"].get():
                text += " The SCF will be spin-unrestricted (UHF) for all molecules."
            else:
                text += (
                    " The SCF will be restricted for closed-shell molecules (RHF) and "
                    "spin-unrestricted (UHF) for all others."
                )

            # MOZYME localized molecular orbitals.
            if ["MOZYME"] == "always" or (
                self._use_mozyme is not None and self._use_mozyme
            ):
                text += (
                    "\n\nThe SCF will be solved using localized molecular orbitals "
                    "(MOZYME), which is faster than the traditional method for larger "
                    "systems."
                )
                used_mozyme = True
            elif self._use_mozyme is not None and not self._use_mozyme:
                used_mozyme = False
            elif P["MOZYME"] == "for larger systems":
                text += (
                    "\n\nThe SCF will be solved using localized molecular orbitals "
                    "(MOZYME) for systems with {nMOZYME} atoms or more. This method is "
                    "faster than the traditional method for larger systems."
                )
                used_mozyme = True
            else:
                used_mozyme = False

            if used_mozyme:
                follow_up = P["MOZYME follow-up"]
                if "exact" in follow_up:
                    text += (
                        " The energy given by MOZYME slowly accumulates error due to "
                        "th increasing non-orthogonality of the localized orbitals "
                        "after many iterations. A single point energy using the "
                        "traditional method will be run to get the correct energy."
                    )
                elif "new" in follow_up:
                    text += (
                        " The energy given by MOZYME slowly accumulates error due to "
                        "the increasing non-orthogonality of the localized orbitals "
                        "after many iterations. A single point energy using fresh "
                        "localized orbitals will be run to get the correct energy."
                    )
                elif follow_up == "none":
                    text += (
                        " The energy given by MOZYME slowly accumulates error due to "
                        "the increasing non-orthogonality of the localized orbitals "
                        "after many iterations. No follow-up calculation will be done, "
                        "so be careful with the final energies produced."
                    )
                    used_mozyme = False
                else:
                    logger.error(f"Don't recognize the MOZYME follow-up: '{follow_up}'")

        # Handle COSMO
        cosmo = P["COSMO"]
        if isinstance(cosmo, bool):
            cosmo = "yes" if cosmo else "no"
        if self.is_expr(cosmo):
            text += (
                "\n\n'{COSMO}' will determine whether to use the COSMO solvation "
                "model. If it is used the parameters will be "
            )
        elif cosmo.lower() == "yes":
            text += "\n\nThe COSMO solvation model will be used with "

        if self.is_expr(cosmo) or cosmo.lower() == "yes":
            text += (
                "dielectric constant = {eps}, solvent radius = {rsolve}, "
                "{nspa} grid points per atom, and a cutoff of {disex}."
            )

        # And bond orders
        if P["bond orders"] == "yes":
            text += "\n\nThe bond orders will be calculated."
        elif P["bond orders"] == "yes, and apply to structure":
            text += (
                "\n\nThe bond orders will be calculated and used to set the bonding "
                "for the structure."
            )

        return self.header + "\n" + __(text, **P, indent=4 * " ").__str__()

    def get_input(self):
        """Get the input for an energy calculation for MOPAC"""
        system, configuration = self.get_system_configuration(None)
        references = self.parent.references

        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )

        # The model chemistry, for labeling properties.
        self.model = P["hamiltonian"]

        # Have to fix formatting for printing...
        PP = dict(P)
        for key in PP:
            if isinstance(PP[key], units_class):
                PP[key] = "{:~P}".format(PP[key])

        if P["MOZYME"] == "always":
            self._use_mozyme = True
        elif (
            P["MOZYME"] == "for larger systems"
            and configuration.n_atoms >= P["nMOZYME"]
        ):
            self._use_mozyme = True
        else:
            self._use_mozyme = False

        # Save the description for later printing
        self.description = []
        self.description.append(__(self.description_text(PP), **PP, indent=4 * " "))

        # Start gathering the keywords
        keywords = copy.deepcopy(P["extra keywords"])
        keywords.append("1SCF")
        keywords.append(P["hamiltonian"])

        if P["hamiltonian"] == "AM1":
            elements = configuration.atoms.symbols
            references.cite(
                raw=self.parent._bibliography["Dewar_1985c"],
                alias="Dewar_1985c",
                module="mopac_step",
                level=1,
                note="Main reference for AM1 + C, H, N, O.",
            )
            for element in ("F", "Cl", "Br", "I"):
                if element in elements:
                    references.cite(
                        raw=self.parent._bibliography["Dewar_1988"],
                        alias="Dewar_1988",
                        module="mopac_step",
                        level=1,
                        note="AM1 parameters for F, Cl, Br, I.",
                    )
                    break
            if "Al" in elements:
                references.cite(
                    raw=self.parent._bibliography["Dewar_1990"],
                    alias="Dewar_1990",
                    module="mopac_step",
                    level=1,
                    note="AM1 parameters for Al.",
                )
            if "Si" in elements:
                references.cite(
                    raw=self.parent._bibliography["Dewar_1987b"],
                    alias="Dewar_1987b",
                    module="mopac_step",
                    level=1,
                    note="AM1 parameters for Si.",
                )
            if "P" in elements:
                references.cite(
                    raw=self.parent._bibliography["Dewar_1989"],
                    alias="Dewar_1989",
                    module="mopac_step",
                    level=1,
                    note="AM1 parameters for P.",
                )
            if "S" in elements:
                references.cite(
                    raw=self.parent._bibliography["Dewar_1990b"],
                    alias="Dewar_1990b",
                    module="mopac_step",
                    level=1,
                    note="AM1 parameters for S.",
                )
            if "Zn" in elements:
                references.cite(
                    raw=self.parent._bibliography["Dewar_1988b"],
                    alias="Dewar_1988b",
                    module="mopac_step",
                    level=1,
                    note="AM1 parameters for Zn.",
                )
            if "Ge" in elements:
                references.cite(
                    raw=self.parent._bibliography["Dewar_1989b"],
                    alias="Dewar_1989b",
                    module="mopac_step",
                    level=1,
                    note="AM1 parameters for Ge.",
                )
            if "Mo" in elements:
                references.cite(
                    raw=self.parent._bibliography["Voityuk_2000"],
                    alias="Voityuk_2000",
                    module="mopac_step",
                    level=1,
                    note="AM1 parameters for Mo.",
                )
            if "Hg" in elements:
                references.cite(
                    raw=self.parent._bibliography["Dewar_1989c"],
                    alias="Dewar_1989c",
                    module="mopac_step",
                    level=1,
                    note="AM1 parameters for Hg.",
                )
            for element in (
                "Li",
                "Be",
                "Na",
                "Mg",
                "K",
                "Ca",
                "Ga",
                "As",
                "Se",
                "Rb",
                "Sr",
                "In",
                "Sn",
                "Sb",
                "Te",
                "Cs",
                "Ba",
                "Pb",
                "Bi",
            ):
                if element in elements:
                    references.cite(
                        raw=self.parent._bibliography["Stewart_2004"],
                        alias="Stewart_2004",
                        module="mopac_step",
                        level=1,
                        note="AM1 parameterization for main-group elements.",
                    )
                    break
        elif P["hamiltonian"] == "MNDO" or P["hamiltonian"] == "MNDOD":
            elements = configuration.atoms.symbols
            references.cite(
                raw=self.parent._bibliography["Dewar_1977"],
                alias="Dewar_1977",
                module="mopac_step",
                level=1,
                note="Main reference for MNDO + C, H, N, O.",
            )
            if "Be" in elements:
                references.cite(
                    raw=self.parent._bibliography["Dewar_1978"],
                    alias="Dewar_1978",
                    module="mopac_step",
                    level=1,
                    note="MNDO parameters for Be.",
                )
            if "B" in elements or "Al" in elements:
                if "B" in elements or P["hamiltonian"] == "MNDO":
                    references.cite(
                        raw=self.parent._bibliography["Davis_1981"],
                        alias="Davis_1981",
                        module="mopac_step",
                        level=1,
                        note="MNDO parameters for B and Al.",
                    )
            if "F" in elements:
                references.cite(
                    raw=self.parent._bibliography["Dewar_1978b"],
                    alias="Dewar_1978b",
                    module="mopac_step",
                    level=1,
                    note="MNDO parameters for F.",
                )
            if "Si" in elements and P["hamiltonian"] == "MNDO":
                references.cite(
                    raw=self.parent._bibliography["Dewar_1986"],
                    alias="Dewar_1986",
                    module="mopac_step",
                    level=1,
                    note="Revised MNDO parameters for Si.",
                )
            if "P" in elements and P["hamiltonian"] == "MNDO":
                references.cite(
                    raw=self.parent._bibliography["Dewar_1978b"],
                    alias="Dewar_1978b",
                    module="mopac_step",
                    level=1,
                    note="MNDO parameters for P.",
                )
            if "S" in elements and P["hamiltonian"] == "MNDO":
                references.cite(
                    raw=self.parent._bibliography["Dewar_1986b"],
                    alias="Dewar_1986b",
                    module="mopac_step",
                    level=1,
                    note="MNDO parameters for S.",
                )
            if "Cl" in elements and P["hamiltonian"] == "MNDO":
                references.cite(
                    raw=self.parent._bibliography["Dewar_1983"],
                    alias="Dewar_1983",
                    module="mopac_step",
                    level=1,
                    note="MNDO parameters for Cl.",
                )
            if "Zn" in elements and P["hamiltonian"] == "MNDO":
                references.cite(
                    raw=self.parent._bibliography["Dewar_1986c"],
                    alias="Dewar_1986c",
                    module="mopac_step",
                    level=1,
                    note="MNDO parameters for Zn.",
                )
            if "Ge" in elements:
                references.cite(
                    raw=self.parent._bibliography["Dewar_1987"],
                    alias="Dewar_1987",
                    module="mopac_step",
                    level=1,
                    note="MNDO parameters for Ge.",
                )
            if "Br" in elements and P["hamiltonian"] == "MNDO":
                references.cite(
                    raw=self.parent._bibliography["Dewar_1983b"],
                    alias="Dewar_1983b",
                    module="mopac_step",
                    level=1,
                    note="MNDO parameters for Br.",
                )
            if "Sn" in elements:
                references.cite(
                    raw=self.parent._bibliography["Dewar_1984"],
                    alias="Dewar_1984",
                    module="mopac_step",
                    level=1,
                    note="MNDO parameters for Sn.",
                )
            if "I" in elements and P["hamiltonian"] == "MNDO":
                references.cite(
                    raw=self.parent._bibliography["Dewar_1984b"],
                    alias="Dewar_1984b",
                    module="mopac_step",
                    level=1,
                    note="MNDO parameters for I.",
                )
            if "Hg" in elements and P["hamiltonian"] == "MNDO":
                references.cite(
                    raw=self.parent._bibliography["Dewar_1985"],
                    alias="Dewar_1985",
                    module="mopac_step",
                    level=1,
                    note="MNDO parameters for Hg.",
                )
            if "Pb" in elements:
                references.cite(
                    raw=self.parent._bibliography["Dewar_1985b"],
                    alias="Dewar_1985b",
                    module="mopac_step",
                    level=1,
                    note="MNDO parameters for Pb.",
                )
            for element in (
                "Na",
                "Mg",
                "K",
                "Ca",
                "Ga",
                "As",
                "Se",
                "Rb",
                "Sr",
                "In",
                "Sb",
                "Te",
                "Cs",
                "Ba",
                "Tl",
                "Bi",
            ):
                if element in elements:
                    references.cite(
                        raw=self.parent._bibliography["Stewart_2004"],
                        alias="Stewart_2004",
                        module="mopac_step",
                        level=1,
                        note="MNDO parameterization for main-group elements.",
                    )
                    break
            if P["hamiltonian"] == "MNDOD":
                for element in (
                    "Al",
                    "Si",
                    "P",
                    "S",
                    "Cl",
                    "Br",
                    "I",
                    "Zn",
                    "Cd",
                    "Hg",
                ):
                    if element in elements:
                        references.cite(
                            raw=self.parent._bibliography["Thiel_1992"],
                            alias="Thiel_1992",
                            module="mopac_step",
                            level=1,
                            note=("MNDO-D formalism for d-orbitals."),
                        )
                        references.cite(
                            raw=self.parent._bibliography["Thiel_1996"],
                            alias="Thiel_1996",
                            module="mopac_step",
                            level=1,
                            note=(
                                "MNDO-D, parameters for Al, Si, P, S, Cl, Br, "
                                "I, Zn, Cd, and Hg."
                            ),
                        )
                        break
        elif P["hamiltonian"] == "PM3":
            elements = configuration.atoms.symbols
            references.cite(
                raw=self.parent._bibliography["Stewart_1989"],
                alias="Stewart_1989",
                module="mopac_step",
                level=1,
                note="The citation for the MOPAC parameterization.",
            )
            for element in (
                "Be",
                "Mg",
                "Zn",
                "Ga",
                "Ge",
                "As",
                "Se",
                "Cd",
                "In",
                "Sn",
                "Sb",
                "Te",
                "Hg",
                "Tl",
                "Pb",
                "Bi",
            ):
                if element in elements:
                    references.cite(
                        raw=self.parent._bibliography["Stewart_1991"],
                        alias="Stewart_1991",
                        module="mopac_step",
                        level=1,
                        note="The citation for the MOPAC parameterization.",
                    )
                    break
            if "Li" in elements:
                references.cite(
                    raw=self.parent._bibliography["Anders_1993"],
                    alias="Anders_1993",
                    module="mopac_step",
                    level=1,
                    note="The citation for the MOPAC parameterization.",
                )
            for element in ("B", "Na", "K", "Ca", "Rb", "Sr", "Cs", "Ba"):
                if element in elements:
                    references.cite(
                        raw=self.parent._bibliography["Stewart_2004"],
                        alias="Stewart_2004",
                        module="mopac_step",
                        level=1,
                        note="The citation for the MOPAC parameterization.",
                    )
                    break
        elif P["hamiltonian"].lower() == "pm6-org":
            references.cite(
                raw=self.parent._bibliography["PM6-ORG"],
                alias="PM6-ORG",
                module="mopac_step",
                level=1,
                note="The PM6-ORG parameterization in MOPAC.",
            )
        elif "PM6" in P["hamiltonian"]:
            references.cite(
                raw=self.parent._bibliography["Stewart_2007"],
                alias="Stewart_2007",
                module="mopac_step",
                level=1,
                note="The PM6 parameterization in MOPAC.",
            )
            if P["hamiltonian"] == "PM6-D3":
                references.cite(
                    raw=self.parent._bibliography["Grimme_2010"],
                    alias="Grimme_2010",
                    module="mopac_step",
                    level=1,
                    note="Dispersion correction by Grimme, et al.",
                )
            if P["hamiltonian"] == "PM6-DH+":
                references.cite(
                    raw=self.parent._bibliography["Korth_2010"],
                    alias="Korth_2010",
                    module="mopac_step",
                    level=1,
                    note="Hydrogen-bonding correction by Korth.",
                )
            if "PM6-DH2" in P["hamiltonian"]:
                references.cite(
                    raw=self.parent._bibliography["Korth_2009"],
                    alias="Korth_2009",
                    module="mopac_step",
                    level=1,
                    note="Hydrogen-bonding and dispersion correction.",
                )
                references.cite(
                    raw=self.parent._bibliography["Rezac_2009"],
                    alias="Rezac_2009",
                    module="mopac_step",
                    level=1,
                    note="Hydrogen-bonding and dispersion correction.",
                )
            if P["hamiltonian"] == "PM6-DH2x":
                references.cite(
                    raw=self.parent._bibliography["Rezac_2011"],
                    alias="Rezac_2011",
                    module="mopac_step",
                    level=1,
                    note="Halogen-bonding correction.",
                )
            if "PM6-D3H4" in P["hamiltonian"]:
                references.cite(
                    raw=self.parent._bibliography["Rezac_2011"],
                    alias="Rezac_2011",
                    module="mopac_step",
                    level=1,
                    note="Hydrogen-bonding and dispersion correction.",
                )
                references.cite(
                    raw=self.parent._bibliography["Vorlova_2015"],
                    alias="Vorlova_2015",
                    module="mopac_step",
                    level=1,
                    note="Hydrogen-hydrogen repulsion correction.",
                )
            if P["hamiltonian"] == "PM6-D3H4x":
                references.cite(
                    raw=self.parent._bibliography["Brahmkshatriya_2013"],
                    alias="Brahmkshatriya_2013",
                    module="mopac_step",
                    level=1,
                    note="Halogen-oxygen and halogen-nitrogen correction.",
                )
        elif "PM7" in P["hamiltonian"]:
            references.cite(
                raw=self.parent._bibliography["Stewart_2012"],
                alias="Stewart_2012",
                module="mopac_step",
                level=1,
                note="The PM7 parameterization in MOPAC.",
            )
        elif P["hamiltonian"] == "RM1":
            references.cite(
                raw=self.parent._bibliography["Rocha_2006"],
                alias="Rocha_2006",
                module="mopac_step",
                level=1,
                note="RM1 parameterization.",
            )

        # which structure? may need to set default first...
        if P["structure"] == "default":
            if self._id[-1] == "1":
                structure = "initial"
            else:
                structure = "current"
        elif self._id[-1] == "1":
            structure = "initial"
        elif P["structure"] == "current":
            structure = "current"

        if structure == "current":
            keywords.append("OLDGEO")

        if P["convergence"] == "normal":
            pass
        elif P["convergence"] == "precise":
            keywords.append("PRECISE")
        elif P["convergence"] == "relative":
            keywords.append("RELSCF=" + P["relative"])
        elif P["convergence"] == "absolute":
            keywords.append("SCFSCRT=" + P["absolute"])
        else:
            raise RuntimeError(
                "Don't recognize convergence '{}'".format(P["convergence"])
            )

        # The type of calculation
        calculation = P["calculation"].lower()
        if "hf" in calculation:
            if P["uhf"]:
                keywords.append("UHF")

            if P["MOZYME"] == "always":
                keywords.append("MOZYME")
            elif (
                P["MOZYME"] == "for larger systems"
                and configuration.n_atoms >= P["nMOZYME"]
            ):
                keywords.append("MOZYME")
        elif "ci" in calculation:
            # Make sure no UHF specified
            while "UHF" in keywords:
                keywords.remove("UHF")

            # Number of orbitals
            n_orbitals = P["number ci orbitals"]
            n_docc = P["number doubly occupied ci orbitals"]
            if n_docc == "default":
                keywords.append(f"C.I.={n_orbitals}")
            else:
                keywords.append(f"C.I.=({n_orbitals},{n_docc})")

            # Limited CI?
            if "cisdt" in calculation:
                keywords.append("CISDT")
            elif "cisd" in calculation:
                keywords.append("CISD")
            elif "cis" in calculation:
                keywords.append("CIS")
            elif "casci" in calculation:
                pass
            else:
                raise RuntimeError(f"Don't recognize type of CI: '{calculation}'")
            if P["print ci details"]:
                keywords.append("MECI")
        else:
            raise RuntimeError(f"Don't recognize calculation '{calculation}'")

        if P["COSMO"]:
            keywords.append(f"EPS={P['eps']}")
            rsolve = P["rsolve"].to("Å").magnitude
            keywords.append(f"RSOLVE={rsolve}")
            keywords.append(f"NSPA={P['nspa']}")
            keywords.append(f"DISEX={P['disex']}")

        if P["calculate gradients"].lower() != "no":
            keywords.append("GRADIENTS")

        if "yes" in P["bond orders"]:
            keywords.append("BONDS")

        # Add any extra keywords so that they appear at the end
        metadata = self.metadata["keywords"]
        for keyword in P["extra keywords"]:
            if "=" in keyword:
                # Delete the copy already in the list
                keywords.remove(keyword)
                keyword, value = keyword.split("=")
                value = self.parent.get_value(value)
                if keyword not in metadata or "format" not in metadata[keyword]:
                    keywords.append(keyword + "=" + value)
                else:
                    keywords.append(metadata[keyword]["format"].format(keyword, value))

        result = []
        result.append([[*keywords], None, None])

        # Handle MOZYME follow-up calculations
        if "MOZYME" in keywords:
            follow_up = P["MOZYME follow-up"]
            if "exact" in follow_up:
                keywords.remove("MOZYME")
                if "1SCF" not in keywords:
                    keywords.append("1SCF")
                keywords.append("OLDGEO")
                result.append([[*keywords], None, "MOZYME follow-up using MOPAC"])
            elif "new" in follow_up:
                if "1SCF" not in keywords:
                    keywords.append("1SCF")
                keywords.append("OLDGEO")
                result.append([[*keywords], None, "MOZYME follow-up, reinitializing"])
            elif follow_up == "none":
                pass
            else:
                logger.error(f"Don't recognize the MOZYME follow-up: '{follow_up}'")

        # Set the attribute in the main MOPAC step for writing just the input
        self.parent.input_only = P["input only"]

        return result

    def analyze(self, indent="", data_sections=[], out_sections=[], table=None):
        """Parse the output and generating the text output and store the
        data in variables for other stages to access
        """
        options = self.parent.options

        # Get the parameters used
        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )

        system, configuration = self.get_system_configuration(None)

        if P["MOZYME"] == "always":
            used_mozyme = True
        elif (
            P["MOZYME"] == "for larger systems"
            and configuration.n_atoms >= P["nMOZYME"]
        ):
            used_mozyme = True
        else:
            used_mozyme = False

        if used_mozyme:
            follow_up = P["MOZYME follow-up"]
            if "exact" in follow_up:
                pass
            elif "new" in follow_up:
                pass
            elif follow_up == "none":
                used_mozyme = False
            else:
                logger.error(f"Don't recognize the MOZYME follow-up: '{follow_up}'")

        if table is None:
            table = {
                "Property": [],
                "Value": [],
                "Units": [],
            }
        text = ""

        if used_mozyme:
            data = {**data_sections[0]}
            data.update(data_sections[1])
        else:
            data = data_sections[0]

        if "POINT_GROUP" in data and data["POINT_GROUP"] != "":
            table["Property"].append("Symmetry")
            table["Value"].append(data["POINT_GROUP"])
            table["Units"].append("")
        else:
            table["Property"].append("Symmetry")
            table["Value"].append("?")
            table["Units"].append("")

        if "HEAT_OF_FORMATION" in data:
            tmp = data["HEAT_OF_FORMATION"]
            table["Property"].append("Enthalpy of Formation")
            table["Value"].append(f"{tmp:.2f}")
            table["Units"].append("kcal/mol")

            tmp = Q_(tmp, "kcal/mol").to("kJ/mol").magnitude
            data["Enthalpy of Formation"] = tmp
            table["Property"].append("")
            table["Value"].append(f"{tmp:.2f}")
            table["Units"].append("kJ/mol")

            if used_mozyme and "HEAT_OF_FORMATION" in data_sections[0]:
                tmp2 = data_sections[0]["HEAT_OF_FORMATION"]
                tmp2 = Q_(tmp2, "kcal/mol").to("kJ/mol").magnitude
                table["Property"].append("MOZYME Enthalpy of Formation")
                table["Value"].append(f"{tmp2:.2f}")
                table["Units"].append("kJ/mol")

                tmp2 -= tmp
                table["Property"].append("Othonormality error in EoF")
                table["Value"].append(f"{tmp2:.2f}")
                table["Units"].append("kJ/mol")

                if abs(tmp2) > 10.0:
                    text += (
                        "\n\nWarning: The non-orthogonality of the localized orbitals "
                        f"led to an error in the enthalpy of formation of {tmp2:.2f} "
                        "kJ/mol. This is expected if there were many iterations of "
                        "geometry optimization. Otherwise check the results carefully."
                    )
                    if "exact" in follow_up:
                        text += (
                            " You followed up with an exact calculation. A large "
                            "difference in the energy could indicate a problem "
                            "with the localized molecular orbitals. Check the MOPAC "
                            "output carefully!"
                        )

        if "GRADIENT_NORM" in data:
            tmp = data["GRADIENT_NORM"]
            table["Property"].append("Gradient Norm")
            table["Value"].append(f"{tmp:.2f}")
            table["Units"].append("kcal/mol/Å")

        if "RMSD" in data:
            tmp = data["RMSD"]
            table["Property"].append("RMSD in Geometry")
            table["Value"].append(f"{tmp:.2f}")
            table["Units"].append("Å")

        if "maximum displacement" in data:
            tmp = data["maximum displacement"]
            table["Property"].append("Largest Displacement")
            table["Value"].append(f"{tmp:.2f}")
            table["Units"].append("Å")

        if "displaced atom" in data:
            tmp = data["displaced atom"]
            table["Property"].append("Displaced Atom")
            table["Value"].append(f"{tmp + 1}")
            table["Units"].append("")

        if "SPIN_COMPONENT" in data:
            tmp = data["SPIN_COMPONENT"]
            table["Property"].append("Sz")
            table["Value"].append(f"{tmp:.2f}")
            table["Units"].append("")

        if "TOTAL_SPIN" in data:
            tmp = data["TOTAL_SPIN"]
            table["Property"].append("S^2")
            table["Value"].append(f"{tmp:.2f}")
            table["Units"].append("")

        if "IONIZATION_POTENTIAL" in data:
            tmp = data["IONIZATION_POTENTIAL"]
            table["Property"].append("Ionization Energy")
            table["Value"].append(f"{tmp:.2f}")
            table["Units"].append("eV")

        if "DIPOLE" in data:
            tmp = data["DIPOLE"]
            table["Property"].append("Dipole Moment")
            table["Value"].append(f"{tmp:.2f}")
            table["Units"].append("Debye")

        if "EIGENVALUES" in data or "ALPHA_EIGENVALUES" in data:
            Elumo = None
            Ehomo = None
            if "EIGENVALUES" in data and "MOLECULAR_ORBITAL_OCCUPANCIES" in data:
                for occ, E in zip(
                    data["MOLECULAR_ORBITAL_OCCUPANCIES"], data["EIGENVALUES"]
                ):
                    if occ > 0.1:
                        Ehomo = E
                    else:
                        Elumo = E
                        break
            elif (
                "ALPHA_EIGENVALUES" in data
                and "ALPHA_MOLECULAR_ORBITAL_OCCUPANCIES" in data
                and "BETA_EIGENVALUES" in data
                and "BETA_MOLECULAR_ORBITAL_OCCUPANCIES" in data
            ):
                for occ, E in zip(
                    data["ALPHA_MOLECULAR_ORBITAL_OCCUPANCIES"],
                    data["ALPHA_EIGENVALUES"],
                ):
                    if occ > 0.1:
                        Ehomo = E
                    else:
                        Elumo = E
                        break
                for occ, E in zip(
                    data["BETA_MOLECULAR_ORBITAL_OCCUPANCIES"],
                    data["BETA_EIGENVALUES"],
                ):
                    if occ > 0.1:
                        if E > Ehomo:
                            Ehomo = E
                    else:
                        if Elumo is None or E < Elumo:
                            Elumo = E
                        break

            if Ehomo is not None:
                data["HOMO Energy"] = Ehomo
                table["Property"].append("HOMO Energy")
                table["Value"].append(f"{Ehomo:.2f}")
                table["Units"].append("eV")
            if Elumo is not None:
                data["LUMO Energy"] = Elumo
                table["Property"].append("LUMO Energy")
                table["Value"].append(f"{Elumo:.2f}")
                table["Units"].append("eV")
                data["HOMO-LUMO Gap"] = Elumo - Ehomo
                table["Property"].append("Gap")
                table["Value"].append(f"{Elumo - Ehomo:.2f}")
                table["Units"].append("eV")

        if "AREA" in data:
            tmp = data["AREA"]
            table["Property"].append("COSMO Area")
            table["Value"].append(f"{tmp:.2f}")
            table["Units"].append("Å^2")

        if "VOLUME" in data:
            tmp = data["VOLUME"]
            table["Property"].append("COSMO Volume")
            table["Value"].append(f"{tmp:.2f}")
            table["Units"].append("Å^3")

        # For periodic systems the volume and density, cell parameters and stress
        if configuration.periodicity == 3:
            V = configuration.cell.volume
            rho = configuration.density
            a, b, c, alpha, beta, gamma = configuration.cell.parameters
            table["Property"].extend(("𝗮", "𝗯", "𝗰", "𝞪", "𝞫", "𝞬", "V", "ρ"))
            table["Value"].extend(
                (
                    f"{a:.3f}",
                    f"{b:.3f}",
                    f"{c:.3f}",
                    f"{alpha:.1f}",
                    f"{beta:.1f}",
                    f"{gamma:.1f}",
                    f"{V:.1f}",
                    f"{rho:.3f}",
                )
            )
            table["Units"].extend(("Å", "Å", "Å", "°", "°", "°", "Å³", "g/mL"))
            if "VOIGT_STRESS" in data:
                table["Property"].extend(("Sxx", "Syy", "Szz", "Syz", "Sxz", "Sxy"))
                tmp = [f"{float(v):.3f}" for v in data["VOIGT_STRESS"]]
                table["Value"].extend(tmp)
                table["Units"].extend(["GPa"] * 6)

        text_lines = []
        text_lines.append("                     Results")
        text_lines.append(
            tabulate(
                table,
                headers="keys",
                tablefmt="psql",
                colalign=("center", "decimal", "left"),
            )
        )
        text_lines.append("\n\n")

        # Get charges and spins, etc.
        directory = Path(self.directory)
        directory.mkdir(parents=True, exist_ok=True)

        symbols = configuration.atoms.asymmetric_symbols
        atoms = configuration.atoms
        symmetry = configuration.symmetry
        if "ATOM_CHARGES" in data:
            # Add to atoms (in coordinate table)
            if "charge" not in atoms:
                atoms.add_attribute(
                    "charge", coltype="float", configuration_dependent=True
                )
            if symmetry.n_symops == 1:
                chgs = data["ATOM_CHARGES"]
            else:
                chgs, delta = symmetry.symmetrize_atomic_scalar(data["ATOM_CHARGES"])
                delta = np.array(delta)
                max_delta = np.max(abs(delta))
                text_lines.append(
                    "The maximum difference of the charges of symmetry related atoms "
                    f"was {max_delta:.4f}\n"
                )
            atoms["charge"][0:] = chgs

            # Print the charges and dump to a csv file
            chg_tbl = {
                "Atom": [*range(1, len(symbols) + 1)],
                "Element": symbols,
                "Charge": [],
            }
            with open(directory / "atom_properties.csv", "w", newline="") as fd:
                writer = csv.writer(fd)
                if "AO_SPINS" in data:
                    # Sum to atom spins...
                    spins = len(symbols) * [0.0]
                    for spin, indx in zip(data["AO_SPINS"], data["AO_ATOMINDEX"]):
                        spins[indx - 1] += spin

                    # Add to atoms (in coordinate table)
                    if "spin" not in atoms:
                        atoms.add_attribute(
                            "spin", coltype="float", configuration_dependent=True
                        )
                        if symmetry.n_symops == 1:
                            atoms["spin"][0:] = spins
                        else:
                            spins, delta = symmetry.symmetrize_atomic_scalar(spins)
                            atoms["spins"][0:] = spins
                            delta = np.array(delta)
                            max_delta = np.max(abs(delta))
                            text_lines.append(
                                " The maximum difference of the spins of symmetry "
                                f"related atoms was {max_delta:.4f}.\n"
                            )

                    header = "        Atomic charges and spins"
                    chg_tbl["Spin"] = []
                    writer.writerow(["Atom", "Element", "Charge", "Spin"])
                    for atom, symbol, q, s in zip(
                        range(1, len(symbols) + 1),
                        symbols,
                        chgs,
                        spins,
                    ):
                        q = f"{q:.3f}"
                        s = f"{s:.3f}"

                        writer.writerow([atom, symbol, q, s])

                        chg_tbl["Charge"].append(q)
                        chg_tbl["Spin"].append(s)
                else:
                    header = "        Atomic charges"
                    writer.writerow(["Atom", "Element", "Charge"])
                    for atom, symbol, q in zip(
                        range(1, len(symbols) + 1),
                        symbols,
                        chgs,
                    ):
                        q = f"{q:.2f}"
                        writer.writerow([atom, symbol, q])

                        chg_tbl["Charge"].append(q)
            if len(symbols) <= int(options["max_atoms_to_print"]):
                text_lines.append(header)
                text_lines.append(
                    tabulate(
                        chg_tbl,
                        headers="keys",
                        tablefmt="psql",
                        colalign=("center", "center"),
                    )
                )

        text = str(__(text, **data, indent=8 * " "))
        text += "\n\n"
        text += textwrap.indent("\n".join(text_lines), 12 * " ")

        if "BOND_ORDERS" in data:
            text += self._bond_orders(
                P["bond orders"], data["BOND_ORDERS"], configuration
            )

        printer.normal(text)

        if used_mozyme and "CPU_TIME" in data_sections[0]:
            t0 = data_sections[0]["CPU_TIME"]
            if "CPU_TIME" in data:
                t1 = data["CPU_TIME"]
                text = (
                    f"This MOZYME calculation took {t0:.2f} s and the follow-up "
                    f"took {t1:.2f} s."
                )
            else:
                text = f"The MOZYME calculation took {t0:.2f} s."
        elif "CPU_TIME" in data:
            t0 = data_sections[0]["CPU_TIME"]
            text = f"This calculation took {t0:.2f} s."
        printer.normal(str(__(text, **data, indent=4 * " ")))

        # Put any requested results into variables or tables
        if "HEAT_OF_FORMATION" in data:
            data["energy"] = data["HEAT_OF_FORMATION"]
        else:
            # See if it is in the output. PM7-TS does not write to AUX!
            lines = iter(out_sections[0])
            for line in lines:
                if "FINAL HEAT OF FORMATION =" in line:
                    try:
                        data["HEAT_OF_FORMATION"] = float(line.split()[5])
                        data["energy"] = data["HEAT_OF_FORMATION"]
                    except Exception:
                        pass

        if "GRADIENTS" in data:
            # Ouch. MOPAC has the gradients of the cell vectors at the end
            n_atoms = configuration.n_atoms
            tmp = np.array(data["GRADIENTS"][: 3 * n_atoms]).reshape(n_atoms, 3)
            # Remove any translation
            delta = np.average(tmp, axis=0)
            tmp -= delta
            del data["GRADIENTS"]
            data["gradients"] = tmp.tolist()
            if "save" in P["calculate gradients"].lower():
                # MOPAC has Cartesian gradients, in kcal/mol/Å
                factor = Q_("kcal/mol/Å").m_as("kJ/mol/Å")
                tmp *= factor
                configuration.atoms.set_gradients(tmp, fractionals=False)
        elif "GRADIENT_NORM" in data:
            # MOPAC does not currently write gradients to the AUX file if they are small
            # They are, however, written to the output file.
            lines = iter(out_sections[0])
            for line in lines:
                if "FINAL  POINT  AND  DERIVATIVES" in line:
                    for line in lines:
                        if "GRADIENT" in line:
                            break
                    tmp = []
                    for line in lines:
                        if "KCAL/ANGSTROM" not in line:
                            break
                        tmp.append(float(line.split()[5]))
                    if len(tmp) == 3 * configuration.n_atoms:
                        tmp = np.array(tmp).reshape(-1, 3)
                        # Remove any translation
                        delta = np.average(tmp, axis=0)
                        tmp -= delta
                        data["gradients"] = tmp.tolist()

        # Handle the force constant matrix (Hessian) if it exists
        if "HESSIAN_MATRIX" in data:
            # It is mass weighted so we need to remove the weighting
            if "ISOTOPIC_MASSES" not in data:
                raise RuntimeError("Found no atomic masses")
            # Expand the mass array for x, y, z
            mass = [v for v in data["ISOTOPIC_MASSES"] for j in range(3)]

            # Get the atom part of the force constant matrix.
            hessian = data["HESSIAN_MATRIX"]
            ij = 0
            factor = Q_(1.0, "mdyne/Å").m_as("kcal/mol/Å^2")
            tmp = []
            for i, mass_i in enumerate(mass):
                for mass_j in mass[0 : i + 1]:
                    tmp.append(factor * hessian[ij] * sqrt(mass_i * mass_j))
                    ij += 1
            data["force constants"] = tmp

        self.store_results(
            configuration=configuration,
            data=data,
            create_tables=self.parameters["create tables"].get(),
        )

    def _bond_orders(self, control, bond_order_matrix, configuration):
        """Analyze and print the bond orders, and optionally use for the bonding
        in the structure.

        Parameters
        ----------
        control : str
            The control option for the bond order analysis
        bond_order_matrix : [float]
            Lower triangular part of the bond order matrix.
        configuration : molsystem.Configuration
            The configuration to put the bonds on, if requested.
        """
        text = ""
        n_atoms = configuration.n_atoms

        if n_atoms == 1:
            return "\n\n        No bonds, since there is only one atom.\n"

        bond_i = []
        bond_j = []
        bond_order = []
        bond_order_str = []
        orders = []
        ij = 0
        for j in range(n_atoms):
            for i in range(j + 1):
                if i != j:
                    order = bond_order_matrix[ij]
                    if order > 0.5:
                        bond_i.append(i)
                        bond_j.append(j)
                        if order > 1.3 and order < 1.7:
                            bond_order.append(5)
                            bond_order_str.append("aromatic")
                        else:
                            bond_order.append(round(order))
                            bond_order_str.append(str(round(order)))
                        orders.append(order)
                ij += 1

        if len(bond_order) > 0:
            symbols = configuration.atoms.symbols
            options = self.parent.options
            text_lines = []
            if len(symbols) <= int(options["max_atoms_to_print"]):
                name = configuration.atoms.names
                table = {
                    "i": [name[i] for i in bond_i],
                    "j": [name[j] for j in bond_j],
                    "bond order": [f"{o:6.3f}" for o in orders],
                    "bond multiplicity": bond_order_str,
                }
                tmp = tabulate(
                    table,
                    headers="keys",
                    tablefmt="psql",
                    disable_numparse=True,
                    colalign=("center", "center", "center", "center"),
                )
                length = len(tmp.splitlines()[0])
                text_lines.append("\n")
                text_lines.append("Bond Orders".center(length))
                text_lines.append(tmp)
                text += "\n\n"
                text += textwrap.indent("\n".join(text_lines), 12 * " ")

            if control == "yes, and apply to structure":
                ids = configuration.atoms.ids
                iatoms = [ids[i] for i in bond_i]
                jatoms = [ids[j] for j in bond_j]
                configuration.new_bondset()
                configuration.bonds.append(i=iatoms, j=jatoms, bondorder=bond_order)
                text2 = (
                    "\nReplaced the bonds in the configuration with those from the "
                    "calculated bond orders.\n"
                )

                text += str(__(text2, indent=8 * " "))

        return text



# -*- coding: utf-8 -*-

"""Control parameters for a MOPAC vibrational (IR/Raman) calculation"""

import logging
import mopac_step

logger = logging.getLogger(__name__)


class IRParameters(mopac_step.EnergyParameters):
    """The control parameters for creating a structure from SMILES"""

    parameters = {
        "let": {
            "default": ("no"),
            "kind": ("boolean"),
            "default_units": (""),
            "enumeration": ("yes", "no"),
            "format_string": (""),
            "group": (""),
            "description": ("Use structure even if not a minimum:"),
            "help_text": (
                "Don't stop if the initial structure is not "
                "a minimum or other stationary point."
            ),
        },
    }

    def __init__(self, defaults={}, data=None):
        """Initialize the instance, by default from the default
        parameters given in the class"""

        super().__init__(
            defaults={
                **IRParameters.parameters,
                **defaults,
            },
            data=data,
        )

        # Do any local editing of defaults
        tmp = self["system name"]
        tmp._data["enumeration"] = (*tmp.enumeration, "vibrations with {Hamiltonian}")
        tmp.default = "keep current name"

        tmp = self["configuration name"]
        tmp._data["enumeration"] = ("vibrations with {Hamiltonian}", *tmp.enumeration)
        tmp.default = "vibrations with {Hamiltonian}"



# -*- coding: utf-8 -*-

"""Caculate the forceconstant matrix using MOPAC"""

import logging
from math import sqrt
from pathlib import Path
import textwrap

import mopac_step
import seamm
import seamm_util.printing as printing
from seamm_util import units_class, Q_
from seamm_util.printing import FormattedText as __

logger = logging.getLogger(__name__)
job = printing.getPrinter()
printer = printing.getPrinter("mopac")


class Forceconstants(mopac_step.Energy):
    def __init__(self, flowchart=None, title="Force Constants", extension=None):
        """Initialize the node"""

        logger.debug("Creating Forceconstants {}".format(self))

        super().__init__(flowchart=flowchart, title=title, extension=extension)

        self._calculation = "force constants"
        self._model = None
        self._metadata = mopac_step.metadata
        self.parameters = mopac_step.ForceconstantsParameters()

        self.description = "Force Constants (Hessian) calculation"

    def description_text(self, P=None):
        """Prepare information about what this node will do"""

        if not P:
            P = self.parameters.values_to_dict()

        # The energy part of the description
        tmp = super().description_text(P)
        energy_description = textwrap.dedent("\n".join(tmp.splitlines()[1:]))

        # Put in the description of the energy calculation
        text = "The energy and forces will be c" + energy_description[1:]
        text += "\n\n"

        if P["what"] == "full Hessian":
            text += "Creating the full Hessian"
        elif P["what"] == "atom part only":
            text += "Creating the atom part of the Hessian"
        else:
            text += "Creating the cell part of the Hessian"
        text += " (force constant matrix) for {hamiltonian}"

        if P["what"] == "cell part only":
            if P["two-sided_cell"]:
                text += " using two-sided differences of the strain"
            else:
                text += " using one-sided differences of the strain"
        else:
            if P["two-sided_atoms"]:
                if P["two-sided_cell"]:
                    if P["what"] == "full Hessian":
                        text += (
                            " using two-sided differences for both the atoms and strain"
                        )
                    else:
                        text += " using two-sided differences for the atoms"
                else:
                    if P["what"] == "full Hessian":
                        text += (
                            " using two-sided differences for the atoms and one-sided "
                            " for the strain"
                        )
                    else:
                        text += " using two-sided differences for the atoms"
            else:
                if P["two-sided_cell"]:
                    if P["what"] == "full Hessian":
                        text += (
                            " using one-sided differences for the atoms and two-sided "
                            "for the strain"
                        )
                    else:
                        text += " using one-sided differences for the atoms"
                else:
                    if P["what"] == "full Hessian":
                        text += (
                            " using on-sided differences for both the atoms and strain"
                        )
                    else:
                        text += " using one-sided differences for the atoms"
        text += " with a step size of {stepsize}."
        if P["what"] == "cell part only":
            text += (
                " The strain-strain part of the Hessian will be output in {cell_units}."
            )
        elif P["what"] == "full Hessian":
            text += (
                " The atom-atom and atom-strain part of the Hessian will be output in "
                "{atom_units}, while the the strain-strain part will be output in "
                "{cell_units}."
            )
        else:
            text += " The atom-atom part of the Hessian will be output in {atom_units}."

        return self.header + "\n" + __(text, **P, indent=4 * " ").__str__()

    def get_input(self):
        """Get the input for an optimization MOPAC"""

        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )

        # Have to fix formatting for printing...
        PP = dict(P)
        for key in PP:
            if isinstance(PP[key], units_class):
                PP[key] = "{:~P}".format(PP[key])

        # Save the description for later printing
        self.description = []
        self.description.append(__(self.description_text(PP), **PP, indent=4 * " "))

        _, configuration = self.get_system_configuration(None)

        is_periodic = configuration.periodicity != 0

        inputs = []

        # Get the underlying SCF setup
        original = super().get_input()
        # Have to think about MOZYME
        if len(original) > 1:
            raise NotImplementedError("MOZYME not yet handles in forceconstants")

        keywords, _, _ = original[0]
        if "OLDGEO" in keywords:
            keywords.remove("OLDGEO")
        if "1SCF" not in keywords:
            keywords.append("1SCF")

        # Get the cell vectors
        if is_periodic and P["what"] != "atom part only":
            cell_vectors = configuration.cell.vectors(as_array=False)

            if P["two-sided_cell"]:
                # Loop over the strains creating inputs
                for direction, name in zip(
                    range(6), ("xx", "yy", "zz", "yz", "xz", "xy")
                ):
                    vector = 6 * [0.0]
                    vector[direction] = P["stepsize"]
                    configuration.strain(vector)
                    structure, symlines = self.parent.mopac_structure()
                    inputs.append(
                        [[*keywords], structure, f"Strained +{P['stepsize']} in {name}"]
                    )
                    # Set the cell back to the original cell.
                    configuration.cell.from_vectors(cell_vectors)

                    # Minus displacement
                    vector = 6 * [0.0]
                    vector[direction] = -P["stepsize"]
                    configuration.strain(vector)
                    structure, symlines = self.parent.mopac_structure()
                    inputs.append(
                        [[*keywords], structure, f"Strained -{P['stepsize']} in {name}"]
                    )
                    # Set the cell back to the original cell.
                    configuration.cell.from_vectors(cell_vectors)
            else:
                # Original structure
                inputs.append([[*keywords], None, None])

                # Loop over the strains creating inputs
                for direction, name in zip(
                    range(6), ("xx", "yy", "zz", "yz", "xz", "xy")
                ):
                    vector = 6 * [0.0]
                    vector[direction] = P["stepsize"]
                    configuration.strain(vector)
                    structure, symlines = self.parent.mopac_structure()
                    inputs.append(
                        [[*keywords], structure, f"Strained {P['stepsize']} in {name}"]
                    )
                    # Set the cell back to the original cell.
                    configuration.cell.from_vectors(cell_vectors)

        # Add the force calculation
        if P["what"] != "cell part only":
            keywords.remove("1SCF")
            keywords.append("FORCE")
            if "LET" not in keywords:
                keywords.append("LET")
            if "NOREOR" not in keywords:
                keywords.append("NOREOR")
            if P["two-sided_atoms"]:
                if "PRECISE" not in keywords:
                    keywords.append("PRECISE")
            else:
                if "PRECISE" in keywords:
                    keywords.remove("PRECISE")

            structure, symlines = self.parent.mopac_structure()
            inputs.append([[*keywords], structure, "Atomic Hessian calculation"])

        return inputs

    def analyze(self, indent="", data_sections=[], out_sections=[], table=None):
        """Parse the output and generating the text output and store the
        data in variables for other stages to access

        There are 8 calculations:
            SPE at the initial geometry
            6 strains
            The forceconstant calculation

        Since the forceconstant matrix is symmetric (we hope!) we will work with the
        lower triangle as a linear array.
        """
        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )

        _, configuration = self.get_system_configuration(None)
        is_periodic = configuration.periodicity != 0
        n_atoms = configuration.n_atoms
        last = 3 * n_atoms
        result = []

        if P["what"] != "cell part only":
            data = data_sections[-1]
            # It is mass weighted so we need to remove the weighting
            if "ISOTOPIC_MASSES" not in data:
                raise RuntimeError("Found no atomic masses")
            tmp = data["ISOTOPIC_MASSES"]
            n_atoms = len(tmp)

            # Replicate for x, y, z
            mass = []
            for v in tmp:
                for _ in range(3):
                    mass.append(v)

            # Get the atom part of the force constant matrix.
            if "HESSIAN_MATRIX" not in data:
                raise RuntimeError("Found no atomic Hessian matrix!")
            hessian = data["HESSIAN_MATRIX"]

            ij = 0
            factor = Q_(1.0, "mdyne/Å").m_as(P["atom_units"])
            for i in range(last):
                for j in range(i + 1):
                    result.append(factor * hessian[ij] * sqrt(mass[i] * mass[j]))
                    ij += 1

        if is_periodic and P["what"] != "atom part only":
            step = P["stepsize"]
            if P["two-sided_cell"]:
                # Loop through the strained calculations, using finite-differences
                for strain in range(6):
                    data = data_sections[2 * strain]
                    if "GRADIENTS" not in data:
                        raise RuntimeError(
                            f"Found no gradients for strained structure {strain}."
                        )
                    forces = data["GRADIENTS"]
                    if "TRANS_VECTS" not in data:
                        raise RuntimeError(
                            "Found no translation vectors for strained structure "
                            f"{strain}."
                        )

                    f2 = forces[0:last]
                    x0, y0, z0, x1, y1, z1, x2, y2, z2 = data["TRANS_VECTS"]
                    fx0, fy0, fz0, fx1, fy1, fz1, fx2, fy2, fz2 = forces[last:]
                    V = abs(
                        x2 * (y0 * z1 - z0 * y1)
                        + y2 * (z0 * x1 - x0 * z1)
                        + z2 * (x0 * y1 - y0 * x1)
                    )

                    factor = Q_(1.0 / V, "kcal/mol/Å^3").m_as(P["cell_units"])

                    s2 = []
                    s2.append((x0 * fx0 + y0 * fy0 + z0 * fz0) * factor)
                    s2.append((x1 * fx1 + y1 * fy1 + z1 * fz1) * factor)
                    s2.append((x2 * fx2 + y2 * fy2 + z2 * fz2) * factor)
                    syz = x1 * fx2 + y1 * fy2 + z1 * fz2
                    szy = x2 * fx1 + y2 * fy1 + z2 * fz1
                    s2.append(((syz + szy) / 2) * factor)
                    sxz = x0 * fx2 + y0 * fy2 + z0 * fz2
                    szx = x2 * fx0 + y2 * fy0 + z2 * fz0
                    s2.append(((sxz + szx) / 2) * factor)
                    sxy = x0 * fx1 + y0 * fy1 + z0 * fz1
                    syx = x1 * fx0 + y1 * fy0 + z1 * fz0
                    s2.append(((sxy + syx) / 2) * factor)

                    data = data_sections[2 * strain + 1]
                    if "GRADIENTS" not in data:
                        raise RuntimeError(
                            f"Found no gradients for strained structure -{strain}."
                        )
                    forces = data["GRADIENTS"]
                    if "TRANS_VECTS" not in data:
                        raise RuntimeError(
                            "Found no translation vectors for strained structure "
                            f"-{strain}."
                        )

                    f1 = forces[0:last]
                    x0, y0, z0, x1, y1, z1, x2, y2, z2 = data["TRANS_VECTS"]
                    fx0, fy0, fz0, fx1, fy1, fz1, fx2, fy2, fz2 = forces[last:]
                    V = abs(
                        x2 * (y0 * z1 - z0 * y1)
                        + y2 * (z0 * x1 - x0 * z1)
                        + z2 * (x0 * y1 - y0 * x1)
                    )

                    factor = Q_(1.0 / V, "kcal/mol/Å^3").m_as(P["cell_units"])

                    s1 = []
                    s1.append((x0 * fx0 + y0 * fy0 + z0 * fz0) * factor)
                    s1.append((x1 * fx1 + y1 * fy1 + z1 * fz1) * factor)
                    s1.append((x2 * fx2 + y2 * fy2 + z2 * fz2) * factor)
                    syz = x1 * fx2 + y1 * fy2 + z1 * fz2
                    szy = x2 * fx1 + y2 * fy1 + z2 * fz1
                    s1.append(((syz + szy) / 2) * factor)
                    sxz = x0 * fx2 + y0 * fy2 + z0 * fz2
                    szx = x2 * fx0 + y2 * fy0 + z2 * fz0
                    s1.append(((sxz + szx) / 2) * factor)
                    sxy = x0 * fx1 + y0 * fy1 + z0 * fz1
                    syx = x1 * fx0 + y1 * fy0 + z1 * fz0
                    s1.append(((sxy + syx) / 2) * factor)

                    # atoms
                    if P["what"] == "full Hessian":
                        factor = Q_(1.0, "kcal/mol/Å^2").m_as(P["atom_units"])
                        for v1, v2 in zip(f1, f2):
                            result.append((v2 - v1) / (2 * step) * factor)

                    # strains
                    for i in range(strain + 1):
                        result.append((s2[i] - s1[i]) / (2 * step))
            else:
                # Get the forces for the unperturbed system
                data = data_sections[0]

                if "GRADIENTS" not in data:
                    raise RuntimeError("Found no gradients for unstrained structure.")
                forces = data["GRADIENTS"]
                if "TRANS_VECTS" not in data:
                    raise RuntimeError(
                        "Found no translation vectors for unstrained  structure."
                    )

                f0 = forces[0:last]
                x0, y0, z0, x1, y1, z1, x2, y2, z2 = data["TRANS_VECTS"]
                fx0, fy0, fz0, fx1, fy1, fz1, fx2, fy2, fz2 = forces[last:]
                V = abs(
                    x2 * (y0 * z1 - z0 * y1)
                    + y2 * (z0 * x1 - x0 * z1)
                    + z2 * (x0 * y1 - y0 * x1)
                )

                factor = Q_(1.0 / V, "kcal/mol/Å^3").m_as(P["cell_units"])

                s0 = []
                s0.append((x0 * fx0 + y0 * fy0 + z0 * fz0) * factor)
                s0.append((x1 * fx1 + y1 * fy1 + z1 * fz1) * factor)
                s0.append((x2 * fx2 + y2 * fy2 + z2 * fz2) * factor)
                syz = x1 * fx2 + y1 * fy2 + z1 * fz2
                szy = x2 * fx1 + y2 * fy1 + z2 * fz1
                s0.append(((syz + szy) / 2) * factor)
                sxz = x0 * fx2 + y0 * fy2 + z0 * fz2
                szx = x2 * fx0 + y2 * fy0 + z2 * fz0
                s0.append(((sxz + szx) / 2) * factor)
                sxy = x0 * fx1 + y0 * fy1 + z0 * fz1
                syx = x1 * fx0 + y1 * fy0 + z1 * fz0
                s0.append(((sxy + syx) / 2) * factor)

                # Loop through the strained calculations, using finite-differences
                for strain in range(6):
                    data = data_sections[strain + 1]

                    if "GRADIENTS" not in data:
                        raise RuntimeError(
                            f"Found no gradients for strained structure {strain}."
                        )
                    forces = data["GRADIENTS"]
                    if "TRANS_VECTS" not in data:
                        raise RuntimeError(
                            "Found no translation vectors for strained structure "
                            f"{strain}."
                        )

                    f = forces[0:last]
                    x0, y0, z0, x1, y1, z1, x2, y2, z2 = data["TRANS_VECTS"]
                    fx0, fy0, fz0, fx1, fy1, fz1, fx2, fy2, fz2 = forces[last:]
                    V = abs(
                        x2 * (y0 * z1 - z0 * y1)
                        + y2 * (z0 * x1 - x0 * z1)
                        + z2 * (x0 * y1 - y0 * x1)
                    )

                    factor = Q_(1.0 / V, "kcal/mol/Å^3").m_as(P["cell_units"])

                    s = []
                    s.append((x0 * fx0 + y0 * fy0 + z0 * fz0) * factor)
                    s.append((x1 * fx1 + y1 * fy1 + z1 * fz1) * factor)
                    s.append((x2 * fx2 + y2 * fy2 + z2 * fz2) * factor)
                    syz = x1 * fx2 + y1 * fy2 + z1 * fz2
                    szy = x2 * fx1 + y2 * fy1 + z2 * fz1
                    s.append(((syz + szy) / 2) * factor)
                    sxz = x0 * fx2 + y0 * fy2 + z0 * fz2
                    szx = x2 * fx0 + y2 * fy0 + z2 * fz0
                    s.append(((sxz + szx) / 2) * factor)
                    sxy = x0 * fx1 + y0 * fy1 + z0 * fz1
                    syx = x1 * fx0 + y1 * fy0 + z1 * fz0
                    s.append(((sxy + syx) / 2) * factor)

                    # atoms
                    if P["what"] == "full Hessian":
                        factor = Q_(1.0, "kcal/mol/Å^2").m_as(P["atom_units"])
                        for v, v0 in zip(f, f0):
                            result.append(((v - v0) / step) * factor)

                    # strains
                    for i in range(strain + 1):
                        result.append((s[i] - s0[i]) / step)

        directory = Path(self.directory)
        directory.mkdir(parents=True, exist_ok=True)
        if is_periodic and P["what"] == "cell part only":
            n = 6
        elif not is_periodic or P["what"] == "atom part only":
            n = last
        else:
            n = last + 6

        with open(directory / ".." / "hessian.dat", "w") as fd:
            # Write the header
            fd.write("!molssi hessian 1.0\n")
            if is_periodic and P["what"] == "cell part only":
                fd.write("@cell_dof 6\n")
                fd.write(f"@cell_units {P['cell_units']}\n")
            elif not is_periodic or P["what"] == "atom part only":
                fd.write(f"@atom_dof {last}\n")
                fd.write(f"@atom_units {P['atom_units']}\n")
            else:
                fd.write(f"@atom_dof {last}\n")
                fd.write(f"@atom_units {P['atom_units']}\n")
                fd.write("@cell_dof 6\n")
                fd.write(f"@cell_units {P['cell_units']}\n")
            fd.write(f"@total_dof {n}\n")

            ij = 0
            for i in range(n):
                for j in range(i + 1):
                    fd.write(f"{result[ij]:12.6f} ")
                    ij += 1
                fd.write("\n")

        # Save the force constant matrix (Hessian) in the data sections
        data = data_sections[0]
        fac = Q_(1.0, P["atom_units"]).m_as("kJ/mol/Å^2")
        data["force constants"] = [v * fac for v in result]

        # Let the energy module do its thing
        super().analyze(
            indent=indent,
            data_sections=data_sections,
            out_sections=out_sections,
            table=table,
        )



# -*- coding: utf-8 -*-

"""Installer for the MOPAC plug-in.

This handles any further installation needed after installing the Python
package `mopac-step`.
"""

import logging
from pathlib import Path
import pkg_resources
import subprocess

import seamm_installer

logger = logging.getLogger(__name__)


class Installer(seamm_installer.InstallerBase):
    """Handle further installation needed after installing mopac-step.

    The Python package `MOPAC-step` should already be installed, using `pip`,
    `conda`, or similar. This plug-in-specific installer then checks for the
    MOPAC executable, installing it if needed, and registers its
    location in seamm.ini.

    There are a number of ways to determine which are the correct MOPAC
    executables to use. The aim of this installer is to help the user locate
    the executables. There are a number of possibilities:

    #. The correct executables are already available.

        #. If they are already registered in `seamm.ini` there is nothing else
           to do.

        #. They may be in the current path, in which case they need to be added
           to `seamm.ini`.

        #. If a module system is in use, a module may need to be loaded to give
           access to MOPAC.

        #. They cannot be found automatically, so the user needs to locate the
           executables for the installer.

    #. MOPAC is not installed on the machine. In this case they can be
       installed in a Conda environment. There is one choice:

        #. They can be installed in a separate environment, `seamm-mopac` by
           default.
    """

    def __init__(self, logger=logger):
        # Call the base class initialization, which sets up the commandline
        # parser, amongst other things.
        super().__init__(logger=logger)

        logger.debug("Initializing the MOPAC installer object.")

        # Define this step's details
        self.environment = "seamm-mopac"
        self.section = "mopac-step"
        self.executables = ["mopac"]
        self.resource_path = Path(pkg_resources.resource_filename(__name__, "data/"))

        # The environment.yaml file for Conda installations.
        logger.debug(f"data directory: {self.resource_path}")
        self.environment_file = self.resource_path / "seamm-mopac.yml"

    def exe_version(self, config):
        """Get the version of the MOPAC executable.

        Parameters
        ----------
        path : pathlib.Path
            Path to the executable.

        Returns
        -------
        str
            The version reported by the executable, or 'unknown'.
        """
        environment = config["conda-environment"]
        conda = config["conda"]
        if environment[0] == "~":
            environment = str(Path(environment).expanduser())
            command = f"'{conda}' run --live-stream -p '{environment}' mopac --version"
        elif Path(environment).is_absolute():
            command = f"'{conda}' run --live-stream -p '{environment}' mopac --version"
        else:
            command = f"'{conda}' run --live-stream -n '{environment}' mopac --version"

        logger.debug(f"    Running {command}")
        try:
            result = subprocess.run(
                command,
                stdin=subprocess.DEVNULL,
                capture_output=True,
                text=True,
                shell=True,
            )
        except Exception as e:
            logger.debug(f"    Failed to run {command}: {e}")
            version = "unknown"
        else:
            logger.debug(f"    {result.stdout}")
            version = "unknown"
            lines = result.stdout.splitlines()
            for line in lines:
                line = line.strip()
                tmp = line.split()
                if len(tmp) >= 2:
                    version = tmp[2]
                    break

        return "MOPAC", version
